import asyncio
from typing import (
    Type,
    TypeVar,
    Callable,
    Optional,
    List,
    Tuple,
    Dict,
    Set,
    Sequence,
    Iterator,
    Any,
)
import warnings
import textwrap

import wn
from wn._types import (
    Metadata,
    NormalizeFunction,
    LemmatizeFunction,
)
from wn._util import normalize_form
from wn._db import NON_ROWID
from wn._queries import (
    find_lexicons,
    find_ilis,
    find_proposed_ilis,
    find_entries,
    find_senses,
    find_synsets,
    get_lexicon,
    get_modified,
    get_lexicon_dependencies,
    get_lexicon_extension_bases,
    get_lexicon_extensions,
    get_form_pronunciations,
    get_form_tags,
    get_entry_senses,
    get_sense_relations,
    get_sense_synset_relations,
    get_synset_relations,
    get_synset_members,
    get_synsets_for_ilis,
    get_examples,
    get_definitions,
    get_syntactic_behaviours,
    get_metadata,
    get_lexicalized,
    get_adjposition,
    get_sense_counts,
    get_lexfile,
    match_for_keyword_in_hypernym_graph,
    refresh_keyword_matching_table,
)
from wn import taxonomy

_INFERRED_SYNSET = "*INFERRED*"


class _DatabaseEntity:
    __slots__ = ("_id",)

    _ENTITY_TYPE = ""

    def __init__(self, _id: int = NON_ROWID):
        self._id = _id  # Database-internal id (e.g., rowid)

    def __eq__(self, other):
        if not isinstance(other, _DatabaseEntity):
            return NotImplemented
        # the _id of different kinds of entities, such as Synset and
        # Sense, can be the same, so make sure they are the same type
        # of object first
        return self._ENTITY_TYPE == other._ENTITY_TYPE and self._id == other._id

    def __lt__(self, other):
        if not isinstance(other, _DatabaseEntity):
            return NotImplemented
        elif self._ENTITY_TYPE != other._ENTITY_TYPE:
            return NotImplemented
        else:
            return self._id < other._id

    def __hash__(self):
        return hash((self._ENTITY_TYPE, self._id))


class ILI(_DatabaseEntity):
    """A class for interlingual indices."""

    __slots__ = "id", "status", "_definition"
    __module__ = "wn"

    def __init__(
        self,
        id: Optional[str],
        status: str,
        definition: str = None,
        _id: int = NON_ROWID,
    ):
        super().__init__(_id=_id)
        self.id = id
        self.status = status
        self._definition = definition

    def __repr__(self) -> str:
        return f'ILI({repr(self.id) if self.id else "*PROPOSED*"})'

    def definition(self) -> Optional[str]:
        return self._definition

    def metadata(self) -> Metadata:
        """Return the ILI's metadata."""
        table = "proposed_ilis" if self.status == "proposed" else "ilis"
        return get_metadata(self._id, table)


class Lexicon(_DatabaseEntity):
    """A class representing a wordnet lexicon.

    Attributes:
        id: The lexicon's identifier.
        label: The full name of lexicon.
        language: The BCP 47 language code of lexicon.
        email: The email address of the wordnet maintainer.
        license: The URL or name of the wordnet's license.
        version: The version string of the resource.
        url: The project URL of the wordnet.
        citation: The canonical citation for the project.
        logo: A URL or path to a project logo.
    """

    __slots__ = (
        "id",
        "label",
        "language",
        "email",
        "license",
        "version",
        "url",
        "citation",
        "logo",
    )
    __module__ = "wn"

    _ENTITY_TYPE = "lexicons"

    def __init__(
        self,
        id: str,
        label: str,
        language: str,
        email: str,
        license: str,
        version: str,
        url: str = None,
        citation: str = None,
        logo: str = None,
        _id: int = NON_ROWID,
    ):
        super().__init__(_id=_id)
        self.id = id
        self.label = label
        self.language = language
        self.email = email
        self.license = license
        self.version = version
        self.url = url
        self.citation = citation
        self.logo = logo

    def __repr__(self):
        id, ver, lg = self.id, self.version, self.language
        return f"<Lexicon {id}:{ver} [{lg}]>"

    def metadata(self) -> Metadata:
        """Return the lexicon's metadata."""
        return get_metadata(self._id, "lexicons")

    def specifier(self) -> str:
        """Return the *id:version* lexicon specifier."""
        return f"{self.id}:{self.version}"

    def modified(self) -> bool:
        """Return True if the lexicon has local modifications."""
        return get_modified(self._id)

    def requires(self) -> Dict[str, Optional["Lexicon"]]:
        """Return the lexicon dependencies."""
        return dict(
            (f"{id}:{version}", None if _id is None else _to_lexicon(get_lexicon(_id)))
            for id, version, _, _id in get_lexicon_dependencies(self._id)
        )

    def extends(self) -> Optional["Lexicon"]:
        """Return the lexicon this lexicon extends, if any.

        If this lexicon is not an extension, return None.
        """
        bases = get_lexicon_extension_bases(self._id, depth=1)
        if bases:
            return _to_lexicon(get_lexicon(bases[0]))
        return None

    def extensions(self, depth: int = 1) -> List["Lexicon"]:
        """Return the list of lexicons extending this one.

        By default, only direct extensions are included. This is
        controlled by the *depth* parameter, which if you view
        extensions as children in a tree where the current lexicon is
        the root, *depth=1* are the immediate extensions. Increasing
        this number gets extensions of extensions, or setting it to a
        negative number gets all "descendant" extensions.

        """
        return [
            _to_lexicon(get_lexicon(rowid))
            for rowid in get_lexicon_extensions(self._id, depth=depth)
        ]

    def describe(self, full: bool = True) -> str:
        """Return a formatted string describing the lexicon.

        The *full* argument (default: :python:`True`) may be set to
        :python:`False` to omit word and sense counts.

        Also see: :meth:`Wordnet.describe`

        """
        _id = self._id
        substrings: List[str] = [
            f"{self.specifier()}",
            f"  Label  : {self.label}",
            f"  URL    : {self.url}",
            f"  License: {self.license}",
        ]
        if full:
            substrings.extend(
                [
                    f"  Words  : {_desc_counts(find_entries, _id)}",
                    f"  Senses : {sum(1 for _ in find_senses(lexicon_rowids=(_id,)))}",
                ]
            )
        substrings.extend(
            [
                f"  Synsets: {_desc_counts(find_synsets, _id)}",
                f"  ILIs   : "
                f"{sum(1 for ili, *_ in find_ilis(lexicon_rowids=(_id,))):>6}",
            ]
        )
        return "\n".join(substrings)


def _desc_counts(query: Callable, lexid: int) -> str:
    count: Dict[str, int] = {}
    for _, pos, *_ in query(lexicon_rowids=(lexid,)):
        if pos not in count:
            count[pos] = 1
        else:
            count[pos] += 1
    subcounts = ", ".join(f"{pos}: {count[pos]}" for pos in sorted(count))
    return f"{sum(count.values()):>6} ({subcounts})"


class _LexiconElement(_DatabaseEntity):
    __slots__ = "_lexid", "_wordnet"

    def __init__(
        self, _lexid: int = NON_ROWID, _id: int = NON_ROWID, _wordnet: "Wordnet" = None
    ):
        super().__init__(_id=_id)
        self._lexid = _lexid  # Database-internal lexicon id
        if _wordnet is None:
            _wordnet = Wordnet()
        self._wordnet: "Wordnet" = _wordnet

    def lexicon(self):
        return _to_lexicon(get_lexicon(self._lexid))

    def _get_lexicon_ids(self) -> Tuple[int, ...]:
        if self._wordnet._default_mode:
            return tuple(
                {self._lexid}
                | set(get_lexicon_extension_bases(self._lexid))
                | set(get_lexicon_extensions(self._lexid))
            )
        else:
            return self._wordnet._lexicon_ids


class Pronunciation:
    """A class for word form pronunciations."""

    __slots__ = "value", "variety", "notation", "phonemic", "audio"

    def __init__(
        self,
        value: str,
        variety: str = None,
        notation: str = None,
        phonemic: bool = True,
        audio: str = None,
    ):
        self.value = value
        self.variety = variety
        self.notation = notation
        self.phonemic = phonemic
        self.audio = audio


class Tag:
    """A general-purpose tag class for word forms."""

    __slots__ = (
        "tag",
        "category",
    )
    __module__ = "wn"

    def __init__(self, tag: str, category: str):
        self.tag = tag
        self.category = category

    def __eq__(self, other):
        if not isinstance(other, Tag):
            return NotImplemented
        return self.tag == other.tag and self.category == other.category


class Form(str):
    """A word-form string with additional attributes."""

    __slots__ = (
        "_id",
        "id",
        "script",
    )
    __module__ = "wn"

    _id: int
    id: Optional[str]
    script: Optional[str]

    def __new__(
        cls, form: str, id: str = None, script: str = None, _id: int = NON_ROWID
    ):
        obj = str.__new__(cls, form)  # type: ignore
        obj.id = id
        obj.script = script
        obj._id = _id
        return obj

    def __eq__(self, other):
        if isinstance(other, Form) and self.script != other.script:
            return False
        return str.__eq__(self, other)

    def __hash__(self):
        return str.__hash__(self)

    def pronunciations(self) -> List[Pronunciation]:
        return [Pronunciation(*data) for data in get_form_pronunciations(self._id)]

    def tags(self) -> List[Tag]:
        return [Tag(tag, category) for tag, category in get_form_tags(self._id)]


class Word(_LexiconElement):
    """A class for words (also called lexical entries) in a wordnet."""

    __slots__ = "id", "pos", "_forms"
    __module__ = "wn"

    _ENTITY_TYPE = "entries"

    def __init__(
        self,
        id: str,
        pos: str,
        forms: List[Tuple[str, Optional[str], Optional[str], int]],
        _lexid: int = NON_ROWID,
        _id: int = NON_ROWID,
        _wordnet: "Wordnet" = None,
    ):
        super().__init__(_lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self.id = id
        self.pos = pos
        self._forms = forms

    def __repr__(self) -> str:
        return f"Word({self.id!r})"

    def lemma(self) -> Form:
        """Return the canonical form of the word.

        Example:

            >>> wn.words('wolves')[0].lemma()
            'wolf'

        """
        return Form(*self._forms[0])

    def forms(self) -> List[Form]:
        """Return the list of all encoded forms of the word.

        Example:

            >>> wn.words('wolf')[0].forms()
            ['wolf', 'wolves']

        """
        return [Form(*form_data) for form_data in self._forms]

    def senses(self) -> List["Sense"]:
        """Return the list of senses of the word.

        Example:

            >>> wn.words('zygoma')[0].senses()
            [Sense('ewn-zygoma-n-05292350-01')]

        """
        lexids = self._get_lexicon_ids()
        iterable = get_entry_senses(self._id, lexids)
        return [Sense(*sense_data, self._wordnet) for sense_data in iterable]

    def metadata(self) -> Metadata:
        """Return the word's metadata."""
        return get_metadata(self._id, "entries")

    def synsets(self) -> List["Synset"]:
        """Return the list of synsets of the word.

        Example:

            >>> wn.words('addendum')[0].synsets()
            [Synset('ewn-06411274-n')]

        """
        return [sense.synset() for sense in self.senses()]

    def derived_words(self) -> List["Word"]:
        """Return the list of words linked through derivations on the senses.

        Example:

            >>> wn.words('magical')[0].derived_words()
            [Word('ewn-magic-n'), Word('ewn-magic-n')]

        """
        return [
            derived_sense.word()
            for sense in self.senses()
            for derived_sense in sense.get_related("derivation")
        ]

    def translate(
        self,
        lexicon: str = None,
        *,
        lang: str = None,
    ) -> Dict["Sense", List["Word"]]:
        """Return a mapping of word senses to lists of translated words.

        Arguments:
            lexicon: if specified, translate to words in the target lexicon(s)
            lang: if specified, translate to words with the language code

        Example:

            >>> w = wn.words('water bottle', pos='n')[0]
            >>> for sense, words in w.translate(lang='ja').items():
            ...     print(sense, [jw.lemma() for jw in words])
            ...
            Sense('ewn-water_bottle-n-04564934-01') ['水筒']

        """
        result = {}
        for sense in self.senses():
            result[sense] = [
                t_sense.word()
                for t_sense in sense.translate(lang=lang, lexicon=lexicon)
            ]
        return result


T = TypeVar("T", bound="_Relatable")


class _Relatable(_LexiconElement):
    __slots__ = ("id",)

    def __init__(
        self,
        id: str,
        _lexid: int = NON_ROWID,
        _id: int = NON_ROWID,
        _wordnet: "Wordnet" = None,
    ):
        super().__init__(_lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self.id = id

    def relations(self: T, *args: str) -> Dict[str, List[T]]:
        raise NotImplementedError

    def get_related(self: T, *args: str) -> List[T]:
        raise NotImplementedError

    def closure(self: T, *args: str) -> Iterator[T]:
        visited = set()
        queue = self.get_related(*args)
        while queue:
            relatable = queue.pop(0)
            if relatable.id not in visited:
                visited.add(relatable.id)
                yield relatable
                queue.extend(relatable.get_related(*args))

    def relation_paths(self: T, *args: str, end: T = None) -> Iterator[List[T]]:
        agenda: List[Tuple[List[T], Set[T]]] = [
            ([target], {self, target})
            for target in self.get_related(*args)
            if target._id != self._id  # avoid self loops?
        ]
        while agenda:
            path, visited = agenda.pop()
            if end is not None and path[-1] == end:
                yield path
            else:
                related = [
                    target
                    for target in path[-1].get_related(*args)
                    if target not in visited
                ]
                if related:
                    for synset in reversed(related):
                        new_path = list(path) + [synset]
                        new_visited = visited | {synset}
                        agenda.append((new_path, new_visited))
                elif end is None:
                    yield path


class Synset(_Relatable):
    """Class for modeling wordnet synsets."""

    __slots__ = "pos", "_ili"
    __module__ = "wn"

    _ENTITY_TYPE = "synsets"

    def __init__(
        self,
        id: str,
        pos: str,
        ili: str = None,
        _lexid: int = NON_ROWID,
        _id: int = NON_ROWID,
        _wordnet: "Wordnet" = None,
    ):
        super().__init__(id=id, _lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self.pos = pos
        self._ili = ili

    @classmethod
    def empty(
        cls,
        id: str,
        ili: str = None,
        _lexid: int = NON_ROWID,
        _wordnet: "Wordnet" = None,
    ):
        return cls(id, pos="", ili=ili, _lexid=_lexid, _wordnet=_wordnet)

    @property
    def ili(self):
        if self._ili:
            row = next(find_ilis(id=self._ili), None)
        else:
            row = next(find_proposed_ilis(synset_rowid=self._id), None)
        if row is not None:
            return ILI(*row)
        return None

    def __hash__(self):
        # include ili and lexid in the hash so inferred synsets don't
        # hash the same
        return hash((self._ENTITY_TYPE, self._ili, self._lexid, self._id))

    def __repr__(self) -> str:
        return f"Synset({self.id!r})"

    def definition(self) -> Optional[str]:
        """Return the first definition found for the synset.

        Example:

            >>> wn.synsets('cartwheel', pos='n')[0].definition()
            'a wheel that has wooden spokes and a metal rim'

        """
        lexids = self._get_lexicon_ids()
        return next((text for text, _, _, _ in get_definitions(self._id, lexids)), None)

    def examples(self) -> List[str]:
        """Return the list of examples for the synset.

        Example:

            >>> wn.synsets('orbital', pos='a')[0].examples()
            ['"orbital revolution"', '"orbital velocity"']

        """
        lexids = self._get_lexicon_ids()
        exs = get_examples(self._id, "synsets", lexids)
        return [ex for ex, _, _ in exs]

    def senses(self) -> List["Sense"]:
        """Return the list of sense members of the synset.

        Example:

            >>> wn.synsets('umbrella', pos='n')[0].senses()
            [Sense('ewn-umbrella-n-04514450-01')]

        """
        lexids = self._get_lexicon_ids()
        iterable = get_synset_members(self._id, lexids)
        return [Sense(*sense_data, self._wordnet) for sense_data in iterable]

    def lexicalized(self) -> bool:
        """Return True if the synset is lexicalized."""
        return get_lexicalized(self._id, "synsets")

    def lexfile(self) -> Optional[str]:
        """Return the lexicographer file name for this synset, if any."""
        return get_lexfile(self._id)

    def metadata(self) -> Metadata:
        """Return the synset's metadata."""
        return get_metadata(self._id, "synsets")

    def words(self) -> List[Word]:
        """Return the list of words linked by the synset's senses.

        Example:

            >>> wn.synsets('exclusive', pos='n')[0].words()
            [Word('ewn-scoop-n'), Word('ewn-exclusive-n')]

        """
        return [sense.word() for sense in self.senses()]

    def lemmas(self) -> List[Form]:
        """Return the list of lemmas of words for the synset.

        Example:

            >>> wn.synsets('exclusive', pos='n')[0].words()
            ['scoop', 'exclusive']

        """
        return [w.lemma() for w in self.words()]

    def relations(self, *args: str) -> Dict[str, List["Synset"]]:
        """Return a mapping of relation names to lists of synsets.

        One or more relation names may be given as positional
        arguments to restrict the relations returned. If no such
        arguments are given, all relations starting from the synset
        are returned.

        See :meth:`get_related` for getting a flat list of related
        synsets.

        Example:

            >>> button_rels = wn.synsets('button')[0].relations()
            >>> for relname, sslist in button_rels.items():
            ...     print(relname, [ss.lemmas() for ss in sslist])
            ...
            hypernym [['fixing', 'holdfast', 'fastener', 'fastening']]
            hyponym [['coat button'], ['shirt button']]

        """
        d: Dict[str, List["Synset"]] = {}
        for relname, ss in self._get_relations(args):
            if relname in d:
                d[relname].append(ss)
            else:
                d[relname] = [ss]
        return d

    def get_related(self, *args: str) -> List["Synset"]:
        """Return the list of related synsets.

        One or more relation names may be given as positional
        arguments to restrict the relations returned. If no such
        arguments are given, all relations starting from the synset
        are returned.

        This method does not preserve the relation names that lead to
        the related synsets. For a mapping of relation names to
        related synsets, see :meth:`relations`.

        Example:

            >>> fulcrum = wn.synsets('fulcrum')[0]
            >>> [ss.lemmas() for ss in fulcrum.get_related()]
            [['pin', 'pivot'], ['lever']]
        """
        return [ss for _, ss in self._get_relations(args)]

    def _get_relations(self, args: Sequence[str]) -> List[Tuple[str, "Synset"]]:
        targets: List[Tuple[str, "Synset"]] = []

        lexids = self._get_lexicon_ids()

        # first get relations from the current lexicon(s)
        if self._id != NON_ROWID:
            relations = list(get_synset_relations({self._id}, args, lexids))
            targets.extend(
                (row[0], Synset(*row[2:], self._wordnet))
                for row in relations
                if row[5] in lexids
            )

        # then attempt to expand via ILI
        if self._ili is not None and self._wordnet and self._wordnet._expanded_ids:
            expids = self._wordnet._expanded_ids

            # get expanded relation
            expss = find_synsets(ili=self._ili, lexicon_rowids=expids)
            rowids = {rowid for _, _, _, _, rowid in expss} - {self._id, NON_ROWID}
            relations = list(get_synset_relations(rowids, args, expids))
            ilis = {row[4] for row in relations} - {None}

            # map back to target lexicons
            seen = {ss._id for _, ss in targets}
            for row in get_synsets_for_ilis(ilis, lexicon_rowids=lexids):
                if row[-1] not in seen:
                    targets.append((row[0], Synset(*row, self._wordnet)))

            # add empty synsets for ILIs without a target in lexids
            unseen_ilis = ilis - {tgt._ili for _, tgt in targets}
            for rel_row in relations:
                if rel_row[4] in unseen_ilis:
                    ss = Synset.empty(
                        id=_INFERRED_SYNSET,
                        ili=rel_row[4],
                        _lexid=self._lexid,
                        _wordnet=self._wordnet,
                    )
                    targets.append((rel_row[0], ss))

        return targets

    def hypernym_paths(self, simulate_root: bool = False) -> List[List["Synset"]]:
        """Return the list of hypernym paths to a root synset."""
        return taxonomy.hypernym_paths(self, simulate_root=simulate_root)

    def min_depth(self, simulate_root: bool = False) -> int:
        """Return the minimum taxonomy depth of the synset."""
        return taxonomy.min_depth(self, simulate_root=simulate_root)

    def max_depth(self, simulate_root: bool = False) -> int:
        """Return the maximum taxonomy depth of the synset."""
        return taxonomy.max_depth(self, simulate_root=simulate_root)

    def shortest_path(
        self, other: "Synset", simulate_root: bool = False
    ) -> List["Synset"]:
        """Return the shortest path from the synset to the *other* synset."""
        return taxonomy.shortest_path(self, other, simulate_root=simulate_root)

    def common_hypernyms(
        self, other: "Synset", simulate_root: bool = False
    ) -> List["Synset"]:
        """Return the common hypernyms for the current and *other* synsets."""
        return taxonomy.common_hypernyms(self, other, simulate_root=simulate_root)

    def lowest_common_hypernyms(
        self, other: "Synset", simulate_root: bool = False
    ) -> List["Synset"]:
        """Return the common hypernyms furthest from the root."""
        return taxonomy.lowest_common_hypernyms(
            self, other, simulate_root=simulate_root
        )

    def holonyms(self) -> List["Synset"]:
        """Return the list of synsets related by any holonym relation.

        Any of the following relations are traversed: ``holonym``,
        ``holo_location``, ``holo_member``, ``holo_part``,
        ``holo_portion``, ``holo_substance``.

        """
        return self.get_related(
            "holonym",
            "holo_location",
            "holo_member",
            "holo_part",
            "holo_portion",
            "holo_substance",
        )

    def meronyms(self) -> List["Synset"]:
        """Return the list of synsets related by any meronym relation.

        Any of the following relations are traversed: ``meronym``,
        ``mero_location``, ``mero_member``, ``mero_part``,
        ``mero_portion``, ``mero_substance``.

        """
        return self.get_related(
            "meronym",
            "mero_location",
            "mero_member",
            "mero_part",
            "mero_portion",
            "mero_substance",
        )

    def hypernyms(self) -> List["Synset"]:
        """Return the list of synsets related by any hypernym relation.

        Both the ``hypernym`` and ``instance_hypernym`` relations are
        traversed.

        """
        return self.get_related("hypernym", "instance_hypernym")

    def hyponyms(self) -> List["Synset"]:
        """Return the list of synsets related by any hyponym relation.

        Both the ``hyponym`` and ``instance_hyponym`` relations are
        traversed.

        """
        return self.get_related("hyponym", "instance_hyponym")

    def translate(self, lexicon: str = None, *, lang: str = None) -> List["Synset"]:
        """Return a list of translated synsets.

        Arguments:
            lexicon: if specified, translate to synsets in the target lexicon(s)
            lang: if specified, translate to synsets with the language code

        Example:

            >>> es = wn.synsets('araña', lang='es')[0]
            >>> en = es.translate(lexicon='ewn')[0]
            >>> en.lemmas()
            ['spider']

        """
        ili = self._ili
        if not ili:
            return []
        return synsets(ili=ili, lang=lang, lexicon=lexicon)


class Count(int):
    """A count of sense occurrences in some corpus."""

    __module__ = "wn"

    _id: int

    def __new__(cls, value, _id: int = NON_ROWID):
        obj = int.__new__(cls, value)  # type: ignore
        obj._id = _id
        return obj

    def metadata(self) -> Metadata:
        """Return the count's metadata."""
        return get_metadata(self._id, "counts")


class Sense(_Relatable):
    """Class for modeling wordnet senses."""

    __slots__ = "_entry_id", "_synset_id"
    __module__ = "wn"

    _ENTITY_TYPE = "senses"

    def __init__(
        self,
        id: str,
        entry_id: str,
        synset_id: str,
        _lexid: int = NON_ROWID,
        _id: int = NON_ROWID,
        _wordnet: "Wordnet" = None,
    ):
        super().__init__(id=id, _lexid=_lexid, _id=_id, _wordnet=_wordnet)
        self._entry_id = entry_id
        self._synset_id = synset_id

    def __repr__(self) -> str:
        return f"Sense({self.id!r})"

    def word(self) -> Word:
        """Return the word of the sense.

        Example:

            >>> wn.senses('spigot')[0].word()
            Word('pwn-spigot-n')

        """
        return word(id=self._entry_id)

    def synset(self) -> Synset:
        """Return the synset of the sense.

        Example:

            >>> wn.senses('spigot')[0].synset()
            Synset('pwn-03325088-n')

        """
        return self._wordnet.synset(id=self._synset_id)

    def examples(self) -> List[str]:
        """Return the list of examples for the sense."""
        lexids = self._get_lexicon_ids()
        exs = get_examples(self._id, "senses", lexids)
        return [ex for ex, _, _ in exs]

    def lexicalized(self) -> bool:
        """Return True if the sense is lexicalized."""
        return get_lexicalized(self._id, "senses")

    def adjposition(self) -> Optional[str]:
        """Return the adjective position of the sense.

        Values include :python:`"a"` (attributive), :python:`"p"`
        (predicative), and :python:`"ip"` (immediate
        postnominal). Note that this is only relevant for adjectival
        senses. Senses for other parts of speech, or for adjectives
        that are not annotated with this feature, will return
        ``None``.

        """
        return get_adjposition(self._id)

    def frames(self) -> List[str]:
        """Return the list of subcategorization frames for the sense."""
        lexids = self._get_lexicon_ids()
        return get_syntactic_behaviours(self._id, lexids)

    def counts(self) -> List[Count]:
        """Return the corpus counts stored for this sense."""
        lexids = self._get_lexicon_ids()
        return [
            Count(value, _id=_id) for value, _id in get_sense_counts(self._id, lexids)
        ]

    def metadata(self) -> Metadata:
        """Return the sense's metadata."""
        return get_metadata(self._id, "senses")

    def relations(self, *args: str) -> Dict[str, List["Sense"]]:
        """Return a mapping of relation names to lists of senses.

        One or more relation names may be given as positional
        arguments to restrict the relations returned. If no such
        arguments are given, all relations starting from the sense
        are returned.

        See :meth:`get_related` for getting a flat list of related
        senses.

        """
        d: Dict[str, List["Sense"]] = {}
        for relname, s in self._get_relations(args):
            if relname in d:
                d[relname].append(s)
            else:
                d[relname] = [s]
        return d

    def get_related(self, *args: str) -> List["Sense"]:
        """Return a list of related senses.

        One or more relation types should be passed as arguments which
        determine the kind of relations returned.

        Example:

            >>> physics = wn.senses('physics', lexicon='ewn')[0]
            >>> for sense in physics.get_related('has_domain_topic'):
            ...     print(sense.word().lemma())
            ...
            coherent
            chaotic
            incoherent

        """
        return [s for _, s in self._get_relations(args)]

    def _get_relations(self, args: Sequence[str]) -> List[Tuple[str, "Sense"]]:
        lexids = self._get_lexicon_ids()
        iterable = get_sense_relations(self._id, args, lexids)
        return [
            (relname, Sense(sid, eid, ssid, lexid, rowid, self._wordnet))
            for relname, _, sid, eid, ssid, lexid, rowid in iterable
            if lexids is None or lexid in lexids
        ]

    def get_related_synsets(self, *args: str) -> List[Synset]:
        """Return a list of related synsets."""
        lexids = self._get_lexicon_ids()
        iterable = get_sense_synset_relations(self._id, args, lexids)
        return [
            Synset(ssid, pos, ili, lexid, rowid, self._wordnet)
            for _, _, ssid, pos, ili, lexid, rowid in iterable
            if lexids is None or lexid in lexids
        ]

    def translate(self, lexicon: str = None, *, lang: str = None) -> List["Sense"]:
        """Return a list of translated senses.

        Arguments:
            lexicon: if specified, translate to senses in the target lexicon(s)
            lang: if specified, translate to senses with the language code

        Example:

            >>> en = wn.senses('petiole', lang='en')[0]
            >>> pt = en.translate(lang='pt')[0]
            >>> pt.word().lemma()
            'pecíolo'

        """
        synset = self.synset()
        return [
            t_sense
            for t_synset in synset.translate(lang=lang, lexicon=lexicon)
            for t_sense in t_synset.senses()
        ]


# Useful for factory functions of Word, Sense, or Synset
C = TypeVar("C", Word, Sense, Synset)


class Wordnet:
    """Class for interacting with wordnet data.

    A wordnet object acts essentially as a filter by first selecting
    matching lexicons and then searching only within those lexicons
    for later queries. On instantiation, a *lang* argument is a `BCP
    47`_ language code that restricts the selected lexicons to those
    whose language matches the given code. A *lexicon* argument is a
    space-separated list of lexicon specifiers that more directly
    selects lexicons by their ID and version; this is preferable when
    there are multiple lexicons in the same language or multiple
    version with the same ID.

    Some wordnets were created by translating the words from a larger
    wordnet, namely the Princeton WordNet, and then relying on the
    larger wordnet for structural relations. An *expand* argument is a
    second space-separated list of lexicon specifiers which are used
    for traversing relations, but not as the results of
    queries. Setting *expand* to an empty string (:python:`expand=''`)
    disables expand lexicons.

    The *normalizer* argument takes a callable that normalizes word
    forms in order to expand the search. The default function
    downcases the word and removes diacritics via NFKD_ normalization
    so that, for example, searching for *san josé* in the English
    WordNet will find the entry for *San Jose*. Setting *normalizer*
    to :python:`None` disables normalization and forces exact-match
    searching.

    The *lemmatizer* argument may be :python:`None`, which is the
    default and disables lemmatizer-based query expansion, or a
    callable that takes a word form and optional part of speech and
    returns base forms of the original word. To support lemmatizers
    that use the wordnet for instantiation, such as :mod:`wn.morphy`,
    the lemmatizer may be assigned to the :attr:`lemmatizer` attribute
    after creation.

    If the *search_all_forms* argument is :python:`True` (the
    default), searches of word forms consider all forms in the
    lexicon; if :python:`False`, only lemmas are searched. Non-lemma
    forms may include, depending on the lexicon, morphological
    exceptions, alternate scripts or spellings, etc.

    .. _BCP 47: https://en.wikipedia.org/wiki/IETF_language_tag
    .. _NFKD: https://en.wikipedia.org/wiki/Unicode_equivalence#Normal_forms

    Attributes:

        lemmatizer: A lemmatization function or :python:`None`.

    """

    cache: Dict[
        Any,
        Any,
    ] = {}

    @classmethod
    def update_matching_table(cls, keyword: List[str]) -> None:
        if wn.config.match_on_keywords:
            asyncio.get_event_loop().create_task(
                refresh_keyword_matching_table(keyword)
            )

    @classmethod
    def get_keyword_matches(cls, term: str, keyword: List[str] = None) -> Iterator[str]:
        return (x[0] for x in match_for_keyword_in_hypernym_graph(term, keyword))

    @classmethod
    def __getCache(
        cls, *args, **kwargs
    ):
        if (
            args, kwargs
        ) in cls.cache:
            return cls.cache[
                (args, kwargs)
            ]

    __slots__ = (
        "_lexicons",
        "_lexicon_ids",
        "_expanded",
        "_expanded_ids",
        "_default_mode",
        "_normalizer",
        "lemmatizer",
        "_search_all_forms",
    )
    __module__ = "wn"

    # def __new__(
    #     cls,
    #     *args,**kwargs
    # ):
    #     if wn.config.enable_cache:
    #         existing = cls.__getCache(
    #            *args,**kwargs
    #         )
    #         if existing:
    #             return existing
    #         new = super(Wordnet, cls).__new__(cls,*args,**kwargs)
    #     else:
    #         return super(Wordnet, cls).__new__(cls,*args,**kwargs)
    #
    #     return new

    def __init__(
        self,
        lexicon: str = None,
        *,
        lang: str = None,
        expand: str = None,
        normalizer: Optional[NormalizeFunction] = normalize_form,
        lemmatizer: Optional[LemmatizeFunction] = None,
        search_all_forms: bool = True,
    ):
        # if (
        #     lexicon,
        #     lang,
        #     expand,
        #     normalizer,
        #     lemmatizer,
        #     search_all_forms,
        # ) in self.cache:
        #     return
        # self.cache[
        #     (lexicon, lang, expand, normalizer, lemmatizer, search_all_forms)
        # ] = self
        # default mode means any lexicon is searched or expanded upon,
        # but relation traversals only target the source's lexicon
        self._default_mode = not lexicon and not lang

        lexs = list(find_lexicons(lexicon or "*", lang=lang))
        self._lexicons: Tuple[Lexicon, ...] = tuple(map(_to_lexicon, lexs))
        self._lexicon_ids: Tuple[int, ...] = tuple(lx._id for lx in self._lexicons)

        self._expanded: Tuple[Lexicon, ...] = ()
        if expand is None:
            if self._default_mode:
                expand = "*"
            else:
                deps = [
                    (id, ver, _id)
                    for lex in self._lexicons
                    for id, ver, _, _id in get_lexicon_dependencies(lex._id)
                ]
                # warn only if a dep is missing and a lexicon was specified
                if not self._default_mode:
                    missing = " ".join(
                        f"{id}:{ver}" for id, ver, _id in deps if _id is None
                    )
                    if missing:
                        warnings.warn(
                            f"lexicon dependencies not available: {missing}",
                            wn.WnWarning,
                        )
                expand = " ".join(
                    f"{id}:{ver}" for id, ver, _id in deps if _id is not None
                )
        if expand:
            self._expanded = tuple(map(_to_lexicon, find_lexicons(lexicon=expand)))
        self._expanded_ids: Tuple[int, ...] = tuple(lx._id for lx in self._expanded)

        self._normalizer = normalizer
        self.lemmatizer = lemmatizer
        self._search_all_forms = search_all_forms

    def lexicons(self) -> List[Lexicon]:
        """Return the list of lexicons covered by this wordnet."""
        return list(self._lexicons)

    def expanded_lexicons(self) -> List[Lexicon]:
        """Return the list of expand lexicons for this wordnet."""
        return list(self._expanded)

    def word(self, id: str) -> Word:
        """Return the first word in this wordnet with identifier *id*."""
        iterable = find_entries(id=id, lexicon_rowids=self._lexicon_ids)
        try:
            return Word(*next(iterable), self)
        except StopIteration:
            raise wn.Error(f"no such lexical entry: {id}")

    def words(self, form: str = None, pos: str = None) -> List[Word]:
        """Return the list of matching words in this wordnet.

        Without any arguments, this function returns all words in the
        wordnet's selected lexicons. A *form* argument restricts the
        words to those matching the given word form, and *pos*
        restricts words by their part of speech.

        """
        return _find_helper(self, Word, find_entries, form, pos)

    def synset(self, id: str) -> Synset:
        """Return the first synset in this wordnet with identifier *id*."""
        iterable = find_synsets(id=id, lexicon_rowids=self._lexicon_ids)
        try:
            return Synset(*next(iterable), self)
        except StopIteration:
            raise wn.Error(f"no such synset: {id}")

    def synsets(
        self, form: str = None, pos: str = None, ili: str = None
    ) -> List[Synset]:
        """Return the list of matching synsets in this wordnet.

        Without any arguments, this function returns all synsets in
        the wordnet's selected lexicons. A *form* argument restricts
        synsets to those whose member words match the given word
        form. A *pos* argument restricts synsets to those with the
        given part of speech. An *ili* argument restricts synsets to
        those with the given interlingual index; generally this should
        select a unique synset within a single lexicon.

        """
        return _find_helper(self, Synset, find_synsets, form, pos, ili=ili)

    def sense(self, id: str) -> Sense:
        """Return the first sense in this wordnet with identifier *id*."""
        iterable = find_senses(id=id, lexicon_rowids=self._lexicon_ids)
        try:
            return Sense(*next(iterable), self)
        except StopIteration:
            raise wn.Error(f"no such sense: {id}")

    def senses(self, form: str = None, pos: str = None) -> List[Sense]:
        """Return the list of matching senses in this wordnet.

        Without any arguments, this function returns all senses in the
        wordnet's selected lexicons. A *form* argument restricts the
        senses to those whose word matches the given word form, and
        *pos* restricts senses by their word's part of speech.

        """
        return _find_helper(self, Sense, find_senses, form, pos)

    def ili(self, id: str) -> ILI:
        """Return the first ILI in this wordnet with identifer *id*."""
        iterable = find_ilis(id=id, lexicon_rowids=self._lexicon_ids)
        try:
            return ILI(*next(iterable))
        except StopIteration:
            raise wn.Error(f"no such ILI: {id}")

    def ilis(self, status: str = None) -> List[ILI]:
        """Return the list of ILIs in this wordnet.

        If *status* is given, only return ILIs with a matching status.

        """
        iterable = find_ilis(status=status, lexicon_rowids=self._lexicon_ids)
        return [ILI(*ili_data) for ili_data in iterable]

    def describe(self) -> str:
        """Return a formatted string describing the lexicons in this wordnet.

        Example:

            >>> oewn = wn.Wordnet('oewn:2021')
            >>> print(oewn.describe())
            Primary lexicons:
              oewn:2021
                Label  : Open English WordNet
                URL    : https://github.com/globalwordnet/english-wordnet
                License: https://creativecommons.org/licenses/by/4.0/
                Words  : 163161 (a: 8386, n: 123456, r: 4481, s: 15231, v: 11607)
                Senses : 211865
                Synsets: 120039 (a: 7494, n: 84349, r: 3623, s: 10727, v: 13846)
                ILIs   : 120039

        """
        substrings = ["Primary lexicons:"]
        for lex in self.lexicons():
            substrings.append(textwrap.indent(lex.describe(), "  "))
        expanded = self.expanded_lexicons()
        if expanded:
            substrings.append("Expand lexicons:")
            for lex in self.expanded_lexicons():
                substrings.append(textwrap.indent(lex.describe(full=False), "  "))
        return "\n".join(substrings)


def _to_lexicon(data) -> Lexicon:
    rowid, id, label, language, email, license, version, url, citation, logo = data
    return Lexicon(
        id,
        label,
        language,
        email,
        license,
        version,
        url=url,
        citation=citation,
        logo=logo,
        _id=rowid,
    )


def _find_helper(
    w: Wordnet,
    cls: Type[C],
    query_func: Callable,
    form: Optional[str],
    pos: Optional[str],
    ili: str = None,
) -> List[C]:
    """Return the list of matching wordnet entities.

    If the wordnet has a normalizer and the search includes a word
    form, the original word form is searched against both the
    original and normalized columns in the database. Then, if no
    results are found, the search is repeated with the normalized
    form. If the wordnet does not have a normalizer, only exact
    string matches are used.

    """
    kwargs: Dict = {
        "lexicon_rowids": w._lexicon_ids,
        "search_all_forms": w._search_all_forms,
    }
    if ili is not None:
        kwargs["ili"] = ili

    # easy case is when there is no form
    if form is None:
        return [cls(*data, w) for data in query_func(pos=pos, **kwargs)]  # type: ignore

    # if there's a form, we may need to lemmatize and normalize
    lemmatize = w.lemmatizer
    normalize = w._normalizer
    kwargs["normalized"] = bool(normalize)

    forms = lemmatize(form, pos) if lemmatize else {}
    # if no lemmatizer or word not covered by lemmatizer, back off to
    # the original form and pos
    if not forms:
        forms = {pos: {form}}

    # we want unique results here, but a set can make the order
    # erratic, so filter manually later
    results = [
        cls(*data, w)  # type: ignore
        for _pos, _forms in forms.items()
        for data in query_func(forms=_forms, pos=_pos, **kwargs)
    ]
    if not results and normalize:
        results = [
            cls(*data, w)  # type: ignore
            for _pos, _forms in forms.items()
            for data in query_func(
                forms=[normalize(f) for f in _forms], pos=_pos, **kwargs
            )
        ]
    unique_results: List[C] = []
    seen: Set[C] = set()
    for result in results:
        if result not in seen:
            unique_results.append(result)
            seen.add(result)
    return unique_results


def projects() -> List[Dict]:
    """Return the list of indexed projects.

    This returns the same dictionaries of information as
    :meth:`wn.config.get_project_info
    <wn._config.WNConfig.get_project_info>`, but for all indexed
    projects.

    Example:

        >>> infos = wn.projects()
        >>> len(infos)
        36
        >>> infos[0]['label']
        'Open English WordNet'

    """
    index = wn.config.index
    return [
        wn.config.get_project_info(f"{project_id}:{version}")
        for project_id, project_info in index.items()
        for version in project_info.get("versions", [])
        if "resource_urls" in project_info["versions"][version]
    ]


def lexicons(*, lexicon: str = None, lang: str = None) -> List[Lexicon]:
    """Return the lexicons matching a language or lexicon specifier.

    Example:

        >>> wn.lexicons(lang='en')
        [<Lexicon ewn:2020 [en]>, <Lexicon pwn:3.0 [en]>]

    """
    try:
        w = Wordnet(lang=lang, lexicon=lexicon)
    except wn.Error:
        return []
    else:
        return w.lexicons()


def word(id: str, *, lexicon: str = None, lang: str = None) -> Word:
    """Return the word with *id* in *lexicon*.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The *id* argument is then passed to the
    :meth:`Wordnet.word` method.

    >>> wn.word('ewn-cell-n')
    Word('ewn-cell-n')

    """
    return Wordnet(lang=lang, lexicon=lexicon).word(id=id)


def words(
    form: str = None,
    pos: str = None,
    *,
    lexicon: str = None,
    lang: str = None,
) -> List[Word]:
    """Return the list of matching words.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The remaining arguments are passed to the
    :meth:`Wordnet.words` method.

    >>> len(wn.words())
    282902
    >>> len(wn.words(pos='v'))
    34592
    >>> wn.words(form="scurry")
    [Word('ewn-scurry-n'), Word('ewn-scurry-v')]

    """
    return Wordnet(lang=lang, lexicon=lexicon).words(form=form, pos=pos)


def synset(id: str, *, lexicon: str = None, lang: str = None) -> Synset:
    """Return the synset with *id* in *lexicon*.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The *id* argument is then passed to the
    :meth:`Wordnet.synset` method.

    >>> wn.synset('ewn-03311152-n')
    Synset('ewn-03311152-n')

    """
    return Wordnet(lang=lang, lexicon=lexicon).synset(id=id)


def synsets(
    form: str = None,
    pos: str = None,
    ili: str = None,
    *,
    lexicon: str = None,
    lang: str = None,
) -> List[Synset]:
    """Return the list of matching synsets.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The remaining arguments are passed to the
    :meth:`Wordnet.synsets` method.

    >>> len(wn.synsets('couch'))
    4
    >>> wn.synsets('couch', pos='v')
    [Synset('ewn-00983308-v')]

    """
    return Wordnet(lang=lang, lexicon=lexicon).synsets(form=form, pos=pos, ili=ili)


def senses(
    form: str = None,
    pos: str = None,
    *,
    lexicon: str = None,
    lang: str = None,
) -> List[Sense]:
    """Return the list of matching senses.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The remaining arguments are passed to the
    :meth:`Wordnet.senses` method.

    >>> len(wn.senses('twig'))
    3
    >>> wn.senses('twig', pos='n')
    [Sense('ewn-twig-n-13184889-02')]

    """
    return Wordnet(lang=lang, lexicon=lexicon).senses(form=form, pos=pos)


def sense(id: str, *, lexicon: str = None, lang: str = None) -> Sense:
    """Return the sense with *id* in *lexicon*.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The *id* argument is then passed to the
    :meth:`Wordnet.sense` method.

    >>> wn.sense('ewn-flutter-v-01903884-02')
    Sense('ewn-flutter-v-01903884-02')

    """
    return Wordnet(lang=lang, lexicon=lexicon).sense(id=id)


def ili(id: str, *, lexicon: str = None, lang: str = None) -> ILI:
    """Return the interlingual index with *id*.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The *id* argument is then passed to the
    :meth:`Wordnet.ili` method.

    >>> wn.ili(id='i1234')
    ILI('i1234')
    >>> wn.ili(id='i1234').status
    'presupposed'

    """
    return Wordnet(lang=lang, lexicon=lexicon).ili(id=id)


def ilis(
    status: str = None,
    *,
    lexicon: str = None,
    lang: str = None,
) -> List[ILI]:
    """Return the list of matching interlingual indices.

    This will create a :class:`Wordnet` object using the *lang* and
    *lexicon* arguments. The remaining arguments are passed to the
    :meth:`Wordnet.ilis` method.

    >>> len(wn.ilis())
    120071
    >>> len(wn.ilis(status='proposed'))
    2573
    >>> wn.ilis(status='proposed')[-1].definition()
    'the neutrino associated with the tau lepton.'
    >>> len(wn.ilis(lang='de'))
    13818

    """
    return Wordnet(lang=lang, lexicon=lexicon).ilis(status=status)



import pytest

import wn


@pytest.mark.usefixtures("mini_db_1_1")
def test_wordnet_lexicons():
    en = wn.Wordnet("test-en")
    assert len(en.lexicons()) == 1
    assert len(en.expanded_lexicons()) == 0

    en1 = wn.Wordnet("test-en:1")
    assert en.lexicons() == en1.lexicons()
    assert en.expanded_lexicons() == en1.expanded_lexicons()

    en2 = wn.Wordnet(lang="en")
    assert len(en2.lexicons()) == 2
    assert len(en2.expanded_lexicons()) == 0

    es = wn.Wordnet("test-es")
    assert len(es.lexicons()) == 1
    assert len(es.expanded_lexicons()) == 0

    es2 = wn.Wordnet("test-es", expand="test-en")
    assert len(es2.lexicons()) == 1
    assert len(es2.expanded_lexicons()) == 1

    ja = wn.Wordnet("test-ja")
    assert len(ja.lexicons()) == 1
    assert len(ja.expanded_lexicons()) == 1

    ja2 = wn.Wordnet("test-ja", expand="")
    assert len(ja2.lexicons()) == 1
    assert len(ja2.expanded_lexicons()) == 0


@pytest.mark.usefixtures("mini_db")
def test_wordnet_normalize():
    es = wn.Wordnet("test-es")
    assert es.words("Informacion") == es.words("información")
    assert es.words("ínfórmácíón") == es.words("información")
    es = wn.Wordnet("test-es", normalizer=None)
    assert es.words("informacion") == []
    assert es.words("Información") == []

    # The following doesn't necessarily work because any non-None
    # normalizer causes the normalized form column to be tested with
    # the original form
    # es = wn.Wordnet('test-es', normalizer=str.lower)
    # assert es.words('informacion') == []
    # assert es.words('Información') == es.words('información')


@pytest.mark.usefixtures("mini_db")
def test_wordnet_lemmatize():
    # default lemmatizer compares alternative forms
    en = wn.Wordnet("test-en")
    assert en.words("examples") == []
    assert en.words("exemplifying") == en.words("exemplify")
    assert en.words("data") == en.words("datum")

    en = wn.Wordnet("test-en", search_all_forms=False)
    assert en.words("examples") == []
    assert en.words("exemplifying") == []
    assert en.words("data") == []

    def morphy_lite(form, pos):
        result = {pos: {form}}
        if pos in ("n", None) and form.endswith("s"):
            result.setdefault("n", set()).add(form[:-1])
        return result

    en = wn.Wordnet("test-en", lemmatizer=morphy_lite, search_all_forms=False)
    assert en.words("examples", pos="n") == en.words("example")
    assert en.words("examples") == en.words("example")
    assert en.words("exemplifying") == []
    assert en.words("data") == []

    en = wn.Wordnet("test-en", lemmatizer=morphy_lite, search_all_forms=True)
    assert en.words("data") == en.words("datum")
    assert en.words("exemplifying") == en.words("exemplify")



from math import log

import pytest

import wn
from wn.constants import NOUN, VERB, ADJ, ADV
from wn.util import synset_id_formatter
import wn.ic


synset_id = {
    "information": "test-en-0001-n",
    "illustration_example": "test-en-0002-n",
    "sample": "test-en-0004-n",
    "random_sample": "test-en-0005-n",
    "random_sample2": "test-en-0008-n",  # no hypernyms
    "datum": "test-en-0006-n",
    "illustrate_exemplify": "test-en-0003-v",
    "resignate": "test-en-0007-v",
}


words = [
    "For",
    "example",
    ":",
    "random sample",
    ".",
    "This",
    "will",
    "illustrate",
    "and",
    "exemplify",
    ".",
    "A",
    "sample",
    "of",
    "data",
    ".",
]


@pytest.mark.usefixtures("mini_db")
def test_compute_nodistribute_nosmoothing():
    w = wn.Wordnet("test-en:1")
    assert wn.ic.compute(words, w, distribute_weight=False, smoothing=0) == {
        NOUN: {
            synset_id["information"]: 4.0,
            synset_id["illustration_example"]: 3.0,
            synset_id["sample"]: 2.0,
            synset_id["random_sample"]: 1.0,
            synset_id["random_sample2"]: 1.0,
            synset_id["datum"]: 1.0,
            None: 5.0,
        },
        VERB: {
            synset_id["illustrate_exemplify"]: 2.0,
            synset_id["resignate"]: 0.0,
            None: 2.0,
        },
        ADJ: {None: 0.0},
        ADV: {None: 0.0},
    }


@pytest.mark.usefixtures("mini_db")
def test_compute_nodistribute_smoothing():
    w = wn.Wordnet("test-en:1")
    assert wn.ic.compute(words, w, distribute_weight=False, smoothing=1.0) == {
        NOUN: {
            synset_id["information"]: 5.0,
            synset_id["illustration_example"]: 4.0,
            synset_id["sample"]: 3.0,
            synset_id["random_sample"]: 2.0,
            synset_id["random_sample2"]: 2.0,
            synset_id["datum"]: 2.0,
            None: 6.0,
        },
        VERB: {
            synset_id["illustrate_exemplify"]: 3.0,
            synset_id["resignate"]: 1.0,
            None: 3.0,
        },
        ADJ: {None: 1.0},
        ADV: {None: 1.0},
    }


@pytest.mark.usefixtures("mini_db")
def test_compute_distribute_smoothing():
    w = wn.Wordnet("test-en:1")
    assert wn.ic.compute(words, w, distribute_weight=True, smoothing=1.0) == {
        NOUN: {
            synset_id["information"]: 4.5,
            synset_id["illustration_example"]: 3.5,
            synset_id["sample"]: 2.5,
            synset_id["random_sample"]: 1.5,
            synset_id["random_sample2"]: 1.5,
            synset_id["datum"]: 2.0,
            None: 5.0,
        },
        VERB: {
            synset_id["illustrate_exemplify"]: 3.0,
            synset_id["resignate"]: 1.0,
            None: 3.0,
        },
        ADJ: {None: 1.0},
        ADV: {None: 1.0},
    }


@pytest.mark.usefixtures("mini_db")
def test_load(tmp_path):
    w = wn.Wordnet("test-en:1")
    icpath = tmp_path / "foo.dat"
    icpath.write_text(
        "wnver:1234567890AbCdEf\n"
        "1n 4.0 ROOT\n"
        "2n 3.0\n"
        "4n 2.0\n"
        "5n 1.0\n"
        "8n 1.0 ROOT\n"
        "6n 1.0\n"
        "3v 2.0 ROOT\n"
        "7v 0.0 ROOT\n"
    )

    get_synset_id = synset_id_formatter("test-en-{offset:04}-{pos}")
    assert wn.ic.load(icpath, w, get_synset_id=get_synset_id) == wn.ic.compute(
        words, w, distribute_weight=False, smoothing=0.0
    )


@pytest.mark.usefixtures("mini_db")
def test_information_content():
    w = wn.Wordnet("test-en:1")
    ic = wn.ic.compute(words, w)
    info = w.synsets("information")[0]
    samp = w.synsets("sample")[0]
    # info is a root but not the only one, so its IC is not 0.0
    assert wn.ic.information_content(info, ic) == -log(ic["n"][info.id] / ic["n"][None])
    assert wn.ic.information_content(samp, ic) == -log(ic["n"][samp.id] / ic["n"][None])



"""
Local configuration settings.
"""

from typing import Optional, Dict, Sequence, Any
from pathlib import Path

import tomli

from wn import ConfigurationError, ProjectError
from wn._types import AnyPath
from wn.constants import _WORDNET
from wn._util import resources, short_hash

# The directory where downloaded and added data will be stored.
DEFAULT_DATA_DIRECTORY = Path.home() / ".wn_data"
DATABASE_FILENAME = "wn.db"


class WNConfig:
    def __init__(self):
        self._data_directory = DEFAULT_DATA_DIRECTORY
        self._projects = {}
        self._dbpath = self._data_directory / DATABASE_FILENAME
        self.allow_multithreading = False
        self.match_on_keywords = []
        self.enable_cache = False

    @property
    def data_directory(self) -> Path:
        """The file system directory where Wn's data is stored."""
        dir = self._data_directory
        dir.mkdir(exist_ok=True)
        return dir

    @data_directory.setter
    def data_directory(self, path):
        dir = Path(path).expanduser()
        if dir.exists() and not dir.is_dir():
            raise ConfigurationError(f"path exists and is not a directory: {dir}")
        self._data_directory = dir
        self._dbpath = dir / DATABASE_FILENAME

    @property
    def database_path(self):
        """The path to the database file."""
        return self._dbpath

    @property
    def downloads_directory(self):
        """The file system directory where downloads are cached."""
        dir = self.data_directory / "downloads"
        dir.mkdir(exist_ok=True)
        return dir

    @property
    def index(self) -> Dict[str, Dict]:
        """The project index."""
        return self._projects

    def add_project(
        self,
        id: str,
        type: str = _WORDNET,
        label: str = None,
        language: str = None,
        license: str = None,
        error: str = None,
    ) -> None:
        """Add a new wordnet project to the index.

        Arguments:
            id: short identifier of the project
            type: project type (default 'wordnet')
            label: full name of the project
            language: `BCP 47`_ language code of the resource
            license: link or name of the project's default license
            error: if set, the error message to use when the project
              is accessed

        .. _BCP 47: https://en.wikipedia.org/wiki/IETF_language_tag
        """
        if id in self._projects:
            raise ValueError(f"project already added: {id}")
        self._projects[id] = {
            "type": type,
            "label": label,
            "language": language,
            "versions": {},
            "license": license,
        }
        if error:
            self._projects[id]["error"] = error

    def add_project_version(
        self,
        id: str,
        version: str,
        url: str = None,
        error: str = None,
        license: str = None,
    ) -> None:
        """Add a new resource version for a project.

        Exactly one of *url* or *error* must be specified.

        Arguments:
            id: short identifier of the project
            version: version string of the resource
            url: space-separated list of web addresses for the resource
            license: link or name of the resource's license; if not
              given, the project's default license will be used.
            error: if set, the error message to use when the project
              is accessed

        """
        version_data: Dict[str, Any]
        if url and not error:
            version_data = {"resource_urls": url.split()}
        elif error and not url:
            version_data = {"error": error}
        elif url and error:
            raise ConfigurationError(f"{id}:{version} specifies both url and redirect")
        else:
            version_data = {}
        if license:
            version_data["license"] = license
        project = self._projects[id]
        project["versions"][version] = version_data

    def get_project_info(self, arg: str) -> Dict:
        """Return information about an indexed project version.

        If the project has been downloaded and cached, the ``"cache"``
        key will point to the path of the cached file, otherwise its
        value is ``None``.

        Arguments:
            arg: a project specifier

        Example:

            >>> info = wn.config.get_project_info('oewn:2021')
            >>> info['label']
            'Open English WordNet'

        """
        id, _, version = arg.partition(":")
        if id not in self._projects:
            raise ProjectError(f"no such project id: {id}")
        project: Dict = self._projects[id]
        if "error" in project:
            raise ProjectError(project["error"])

        versions: Dict = project["versions"]
        if not version or version == "*":
            version = next(iter(versions), "")
        if not version:
            raise ProjectError(f"no versions available for {id}")
        elif version not in versions:
            raise ProjectError(f"no such version: {version!r} ({id})")
        info = versions[version]
        if "error" in info:
            raise ProjectError(info["error"])

        urls = info.get("resource_urls", [])

        return dict(
            id=id,
            version=version,
            type=project["type"],
            label=project["label"],
            language=project["language"],
            license=info.get("license", project.get("license")),
            resource_urls=urls,
            cache=_get_cache_path_for_urls(self, urls),
        )

    def get_cache_path(self, url: str) -> Path:
        """Return the path for caching *url*.

        Note that in general this is just a path operation and does
        not signify that the file exists in the file system.

        """
        filename = short_hash(url)
        return self.downloads_directory / filename

    def update(self, data: dict) -> None:
        """Update the configuration with items in *data*.

        Items are only inserted or replaced, not deleted. If a project
        index is provided in the ``"index"`` key, then either the
        project must not already be indexed or any project fields
        (label, language, or license) that are specified must be equal
        to the indexed project.

        """
        if "data_directory" in data:
            self.data_directory = data["data_directory"]
        for id, project in data.get("index", {}).items():
            if id in self._projects:
                # validate that they are the same
                _project = self._projects[id]
                for attr in ("label", "language", "license"):
                    if attr in project and project[attr] != _project[attr]:
                        raise ConfigurationError(f"{attr} mismatch for {id}")
            else:
                self.add_project(
                    id,
                    type=project.get("type", _WORDNET),
                    label=project.get("label"),
                    language=project.get("language"),
                    license=project.get("license"),
                    error=project.get("error"),
                )
            for version, info in project.get("versions", {}).items():
                if "url" in info and "error" in project:
                    raise ConfigurationError(
                        f"{id}:{version} url specified with default error"
                    )
                self.add_project_version(
                    id,
                    version,
                    url=info.get("url"),
                    license=info.get("license"),
                    error=info.get("error"),
                )

    def load_index(self, path: AnyPath) -> None:
        """Load and update with the project index at *path*.

        The project index is a TOML_ file containing project and
        version information. For example:

        .. code-block:: toml

           [ewn]
             label = "Open English WordNet"
             language = "en"
             license = "https://creativecommons.org/licenses/by/4.0/"
             [ewn.versions.2019]
               url = "https://en-word.net/static/english-wordnet-2019.xml.gz"
             [ewn.versions.2020]
               url = "https://en-word.net/static/english-wordnet-2020.xml.gz"

        .. _TOML: https://toml.io

        """
        path = Path(path).expanduser()
        with path.open("rb") as indexfile:
            try:
                index = tomli.load(indexfile)
            except tomli.TOMLDecodeError as exc:
                raise ConfigurationError("malformed index file") from exc
        self.update({"index": index})


def _get_cache_path_for_urls(
    config: WNConfig,
    urls: Sequence[str],
) -> Optional[Path]:
    for url in urls:
        path = config.get_cache_path(url)
        if path.is_file():
            return path
    return None


config = WNConfig()
with resources.path("wn", "index.toml") as index_path:
    config.load_index(index_path)



class Error(Exception):
    """Generic error class for invalid wordnet operations."""

    # reset the module so the user sees the public name
    __module__ = "wn"


class DatabaseError(Error):
    """Error class for issues with the database."""

    __module__ = "wn"


class ConfigurationError(Error):
    """Raised on invalid configurations."""

    __module__ = "wn"


class ProjectError(Error):
    """Raised when a project is not found or on errors defined in the index."""

    __module__ = "wn"


class WnWarning(Warning):
    """Generic warning class for dubious worndet operations."""

    # reset the module so the user sees the public name
    __module__ = "wn"



from xml.etree import ElementTree as ET

import pytest

import wn


@pytest.mark.usefixtures("mini_db")
def test_export(mini_lmf_1_0, tmp_path):
    tmpdir = tmp_path / "test_export"
    tmpdir.mkdir()
    tmppath = tmpdir / "mini_lmf_export.xml"
    lexicons = wn.lexicons(lexicon="test-en test-es")
    wn.export(lexicons, tmppath)

    if hasattr(ET, "canonicalize"):  # available from Python 3.8
        # remove comments, indentation, etc.
        orig = ET.canonicalize(from_file=mini_lmf_1_0, strip_text=True)
        temp = ET.canonicalize(from_file=tmppath, strip_text=True)
        # additional transformation to help with debugging
        orig = orig.replace("<", "\n<")
        temp = temp.replace("<", "\n<")
        assert orig == temp



"""
Reader for the Lexical Markup Framework (LMF) format.
"""

from typing import Type, List, Tuple, Dict, Optional, TextIO, BinaryIO, Any, Union, cast
import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict, Literal
else:
    from typing_extensions import TypedDict, Literal
import re
from pathlib import Path
import xml.etree.ElementTree as ET  # for general XML parsing
import xml.parsers.expat  # for fast scanning of Lexicon versions
from xml.sax.saxutils import quoteattr

import wn
from wn._types import AnyPath, VersionInfo
from wn._util import is_xml, version_info
from wn.util import ProgressHandler, ProgressBar


class LMFError(wn.Error):
    """Raised on invalid LMF-XML documents."""


class LMFWarning(Warning):
    """Issued on non-conforming LFM values."""


SUPPORTED_VERSIONS = {"1.0", "1.1"}
_XMLDECL = b'<?xml version="1.0" encoding="UTF-8"?>'
_DOCTYPE = '<!DOCTYPE LexicalResource SYSTEM "{schema}">'
_SCHEMAS = {
    "1.0": "http://globalwordnet.github.io/schemas/WN-LMF-1.0.dtd",
    "1.1": "http://globalwordnet.github.io/schemas/WN-LMF-1.1.dtd",
}
_DOCTYPES = {
    _DOCTYPE.format(schema=schema): version for version, schema in _SCHEMAS.items()
}

_DC_URIS = {
    "1.0": "http://purl.org/dc/elements/1.1/",
    "1.1": "https://globalwordnet.github.io/schemas/dc/",
}
_DC_ATTRS = [
    "contributor",
    "coverage",
    "creator",
    "date",
    "description",
    "format",
    "identifier",
    "publisher",
    "relation",
    "rights",
    "source",
    "subject",
    "title",
    "type",
]
_NS_ATTRS = {
    version: dict(
        [(f"{uri} {attr}", attr) for attr in _DC_ATTRS]
        + [
            ("status", "status"),
            ("note", "note"),
            ("confidenceScore", "confidenceScore"),
        ]
    )
    for version, uri in _DC_URIS.items()
}

_LMF_1_0_ELEMS: Dict[str, str] = {
    "LexicalResource": "lexical-resource",
    "Lexicon": "lexicons",
    "LexicalEntry": "entries",
    "Lemma": "lemma",
    "Form": "forms",
    "Tag": "tags",
    "Sense": "senses",
    "SenseRelation": "relations",
    "Example": "examples",
    "Count": "counts",
    "SyntacticBehaviour": "frames",
    "Synset": "synsets",
    "Definition": "definitions",
    "ILIDefinition": "ili_definition",
    "SynsetRelation": "relations",
}
_LMF_1_1_ELEMS = dict(_LMF_1_0_ELEMS)
_LMF_1_1_ELEMS.update(
    {
        "Requires": "requires",
        "Extends": "extends",
        "Pronunciation": "pronunciations",
        "LexiconExtension": "lexicons",
        "ExternalLexicalEntry": "entries",
        "ExternalLemma": "lemma",
        "ExternalForm": "forms",
        "ExternalSense": "senses",
        "ExternalSynset": "synsets",
    }
)
_VALID_ELEMS = {
    "1.0": _LMF_1_0_ELEMS,
    "1.1": _LMF_1_1_ELEMS,
}
_LIST_ELEMS = {  # elements that collect into lists
    "Lexicon",
    "LexicalEntry",
    "Form",
    "Pronunciation",
    "Tag",
    "Sense",
    "SenseRelation",
    "Example",
    "Count",
    "Synset",
    "Definition",
    "SynsetRelation",
    "SyntacticBehaviour",
    "LexiconExtension",
    "Requires",
    "ExternalLexicalEntry",
    "ExternalForm",
    "ExternalSense",
    "ExternalSynset",
}
_CDATA_ELEMS = {  # elements with inner text
    "Pronunciation",
    "Tag",
    "Definition",
    "ILIDefinition",
    "Example",
    "Count",
}
_META_ELEMS = {  # elements with metadata
    "Lexicon",
    "LexicalEntry",
    "Sense",
    "SenseRelation",
    "Example",
    "Count",
    "Synset",
    "Definition",
    "ILIDefinition",
    "SynsetRelation",
    "LexiconExtension",
}


# WN-LMF Modeling ######################################################

# WN-LMF type-checking is handled via TypedDicts.  Inheritance and
# `total=False` are used to model optionality. For more information
# about this tactic, see https://www.python.org/dev/peps/pep-0589/.


class Metadata(TypedDict, total=False):
    contributor: str
    coverage: str
    creator: str
    date: str
    description: str
    format: str
    identifier: str
    publisher: str
    relation: str
    rights: str
    source: str
    subject: str
    title: str
    type: str
    status: str
    note: str
    confidenceScore: float


_HasId = TypedDict("_HasId", {"id": str})
_HasVersion = TypedDict("_HasVersion", {"version": str})
_HasILI = TypedDict("_HasILI", {"ili": str})
_HasSynset = TypedDict("_HasSynset", {"synset": str})
_MaybeId = TypedDict("_MaybeId", {"id": str}, total=False)
_HasText = TypedDict("_HasText", {"text": str})
_MaybeScript = TypedDict("_MaybeScript", {"script": str}, total=False)
_HasMeta = TypedDict("_HasMeta", {"meta": Optional[Metadata]})
_External = TypedDict("_External", {"external": Literal["true"]})


class ILIDefinition(_HasText, _HasMeta):
    ...


class Definition(_HasText, _HasMeta, total=False):
    language: str
    sourceSense: str


class Relation(_HasMeta):
    target: str
    relType: str


class Example(_HasText, _HasMeta, total=False):
    language: str


class Synset(_HasId, _HasILI, _HasMeta, total=False):
    ili_definition: ILIDefinition
    partOfSpeech: str
    definitions: List[Definition]
    relations: List[Relation]
    examples: List[Example]
    lexicalized: bool
    members: List[str]
    lexfile: str


class ExternalSynset(_HasId, _External, total=False):
    definitions: List[Definition]
    relations: List[Relation]
    examples: List[Example]


class Count(_HasMeta):
    value: int


class Sense(_HasId, _HasSynset, _HasMeta, total=False):
    relations: List[Relation]
    examples: List[Example]
    counts: List[Count]
    lexicalized: bool
    adjposition: str
    subcat: List[str]


class ExternalSense(_HasId, _External, total=False):
    relations: List[Relation]
    examples: List[Example]
    counts: List[Count]


class Pronunciation(_HasText, total=False):
    variety: str
    notation: str
    phonemic: bool
    audio: str


class Tag(_HasText):
    category: str


class _FormChildren(TypedDict, total=False):
    pronunciations: List[Pronunciation]
    tags: List[Tag]


class Lemma(_MaybeScript, _FormChildren):
    writtenForm: str
    partOfSpeech: str


class ExternalLemma(_FormChildren, _External):
    ...


class Form(_MaybeId, _MaybeScript, _FormChildren):
    writtenForm: str


class ExternalForm(_HasId, _FormChildren, _External):
    ...


class _SyntacticBehaviourBase(_MaybeId):
    subcategorizationFrame: str


class SyntacticBehaviour(_SyntacticBehaviourBase, total=False):
    senses: List[str]


class _LexicalEntryBase(_HasId, _HasMeta, total=False):
    forms: List[Form]
    senses: List[Sense]
    frames: List[SyntacticBehaviour]


class LexicalEntry(_LexicalEntryBase):
    lemma: Lemma


class ExternalLexicalEntry(_HasId, _External, total=False):
    lemma: ExternalLemma
    forms: List[Union[Form, ExternalForm]]
    senses: List[Union[Sense, ExternalSense]]


class Dependency(_HasId, _HasVersion, total=False):
    url: str


class _LexiconBase(_HasMeta):
    id: str
    version: str
    label: str
    language: str
    email: str
    license: str


class Lexicon(_LexiconBase, total=False):
    url: str
    citation: str
    logo: str
    requires: List[Dependency]
    entries: List[LexicalEntry]
    synsets: List[Synset]
    frames: List[SyntacticBehaviour]


class LexiconExtension(_LexiconBase, total=False):
    url: str
    citation: str
    logo: str
    extends: Dependency
    requires: List[Dependency]
    entries: List[Union[LexicalEntry, ExternalLexicalEntry]]
    synsets: List[Union[Synset, ExternalSynset]]
    frames: List[SyntacticBehaviour]


class LexicalResource(TypedDict):
    lmf_version: str
    lexicons: List[Union[Lexicon, LexiconExtension]]


# Reading ##############################################################


def is_lmf(source: AnyPath) -> bool:
    """Return True if *source* is a WN-LMF file."""
    source = Path(source).expanduser()
    if not is_xml(source):
        return False
    with source.open(mode="rb") as fh:
        try:
            _read_header(fh)
        except LMFError:
            return False
    return True


def _read_header(fh: BinaryIO) -> str:
    xmldecl = fh.readline().rstrip().replace(b"'", b'"')
    doctype = fh.readline().rstrip().replace(b"'", b'"')

    if xmldecl != _XMLDECL:
        raise LMFError("invalid or missing XML declaration")

    # the XML declaration states that the file is UTF-8 (other
    # encodings are not allowed)
    doctype_decoded = doctype.decode("utf-8")
    if doctype_decoded not in _DOCTYPES:
        raise LMFError("invalid or missing DOCTYPE declaration")

    return _DOCTYPES[doctype_decoded]


def scan_lexicons(source: AnyPath) -> List[Dict]:
    """Scan *source* and return only the top-level lexicon info."""

    source = Path(source).expanduser()
    infos: List[Dict] = []

    lex_re = re.compile(b"<(Lexicon|LexiconExtension|Extends)\\b([^>]*)>", flags=re.M)
    attr_re = re.compile(b"""\\b(id|version|label)=["']([^"']+)["']""", flags=re.M)

    with open(source, "rb") as fh:
        for m in lex_re.finditer(fh.read()):
            lextype, remainder = m.groups()
            info = {
                _m.group(1).decode("utf-8"): _m.group(2).decode("utf-8")
                for _m in attr_re.finditer(remainder)
            }
            if "id" not in info or "version" not in info:
                raise LMFError(f'<{lextype.decode("utf-8")}> missing id or version')
            if lextype != b"Extends":
                infos.append(info)
            elif len(infos) > 0:
                infos[-1]["extends"] = info
            else:
                raise LMFError("invalid use of <Extends> in WN-LMF file")

    return infos


_Elem = Dict[str, Any]  # basic type for the loaded XML data


def load(
    source: AnyPath, progress_handler: Optional[Type[ProgressHandler]] = ProgressBar
) -> LexicalResource:
    """Load wordnets encoded in the WN-LMF format.

    Args:
        source: path to a WN-LMF file
    """
    source = Path(source).expanduser()
    if progress_handler is None:
        progress_handler = ProgressHandler

    version, num_elements = _quick_scan(source)
    progress = progress_handler(
        message="Read", total=num_elements, refresh_interval=10000
    )

    root: Dict[str, _Elem] = {}
    parser = _make_parser(root, version, progress)

    with open(source, "rb") as fh:
        try:
            parser.ParseFile(fh)
        except xml.parsers.expat.ExpatError as exc:
            raise LMFError("invalid or ill-formed XML") from exc

    progress.close()

    resource: LexicalResource = {
        "lmf_version": version,
        "lexicons": [
            _validate(lex) for lex in root["lexical-resource"].get("lexicons", [])
        ],
    }

    return resource


def _quick_scan(source: Path) -> Tuple[str, int]:
    with source.open("rb") as fh:
        version = _read_header(fh)
        # _read_header() only reads the first 2 lines
        remainder = fh.read()
        num_elements = remainder.count(b"</") + remainder.count(b"/>")
    return version, num_elements


def _make_parser(root, version, progress):  # noqa: C901
    stack = [root]
    ELEMS = _VALID_ELEMS[version]
    NS_ATTRS = _NS_ATTRS[version]
    CDATA_ELEMS = _CDATA_ELEMS & set(ELEMS)
    LIST_ELEMS = _LIST_ELEMS & set(ELEMS)

    p = xml.parsers.expat.ParserCreate(namespace_separator=" ")
    has_text = False

    def start(name, attrs):
        nonlocal has_text
        if name in CDATA_ELEMS:
            has_text = True

        if name in _META_ELEMS:
            meta = {}
            for attr in list(attrs):
                if attr in NS_ATTRS:
                    meta[NS_ATTRS[attr]] = attrs.pop(attr)
            attrs["meta"] = meta or None

        if name.startswith("External"):
            attrs["external"] = True

        parent = stack[-1]
        key = ELEMS.get(name)
        if name in LIST_ELEMS:
            parent.setdefault(key, []).append(attrs)
        elif key is None or key in parent:
            raise _unexpected(name, p)
        else:
            parent[key] = attrs

        stack.append(attrs)

    def char_data(data):
        if has_text:
            parent = stack[-1]
            # sometimes the buffering occurs in the middle of text; if
            # so, append the current data
            if "text" in parent:
                parent["text"] += data
            else:
                parent["text"] = data

    def end(name):
        nonlocal has_text
        has_text = False
        stack.pop()
        progress.update(force=(name == "LexicalResource"))

    p.StartElementHandler = start
    p.EndElementHandler = end
    p.CharacterDataHandler = char_data

    return p


def _unexpected(name: str, p: xml.parsers.expat.XMLParserType) -> LMFError:
    return LMFError(f"unexpected element at line {p.CurrentLineNumber}: {name}")


# Validation ###########################################################


def _validate(elem: _Elem) -> Union[Lexicon, LexiconExtension]:
    ext = elem.get("extends")
    if ext:
        assert "id" in ext
        assert "version" in ext
        _validate_lexicon(elem, True)
        return cast(LexiconExtension, elem)
    else:
        _validate_lexicon(elem, False)
        return cast(Lexicon, elem)


def _validate_lexicon(elem: _Elem, extension: bool) -> None:
    for attr in "id", "version", "label", "language", "email", "license":
        assert attr in elem, f"<Lexicon> missing required attribute: {attr}"
    for dep in elem.get("requires", []):
        assert "id" in dep
        assert "version" in dep
    _validate_entries(elem.get("entries", []), extension)
    _validate_synsets(elem.get("synsets", []), extension)
    _validate_frames(elem.get("frames", []))


def _validate_entries(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        assert "id" in elem
        if not extension:
            assert not elem.get("external")
        lemma = elem.get("lemma")
        if not elem.get("external"):
            assert lemma is not None
            elem.setdefault("meta")
        # lemma and forms are the same except for partOfSpeech and id
        if lemma is not None and not lemma.get("external"):
            assert "partOfSpeech" in lemma
        for form in elem.get("forms", []):
            assert not form.get("external") or form.get("id")
        _validate_forms(([lemma] if lemma else []) + elem.get("forms", []), extension)
        _validate_senses(elem.get("senses", []), extension)
        _validate_frames(elem.get("frames", []))


def _validate_forms(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        if not extension:
            assert not elem.get("external")
        if not elem.get("external"):
            assert "writtenForm" in elem
        for pron in elem.get("pronunciations", []):
            pron.setdefault("text", "")
            if pron.get("phonemic"):
                pron["phonemic"] = False if pron["phonemic"] == "false" else True
        for tag in elem.get("tags", []):
            tag.setdefault("text", "")
            assert "category" in tag


def _validate_senses(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        assert "id" in elem
        if not extension:
            assert not elem.get("external")
        if not elem.get("external"):
            assert "synset" in elem
            elem.setdefault("meta")
        for rel in elem.get("relations", []):
            assert "target" in rel
            assert "relType" in rel
            rel.setdefault("meta")
        for ex in elem.get("examples", []):
            ex.setdefault("text", "")
            ex.setdefault("meta")
        for cnt in elem.get("counts", []):
            assert "text" in cnt
            cnt["value"] = int(cnt.pop("text"))
            cnt.setdefault("meta")
        if elem.get("lexicalized"):
            elem["lexicalized"] = False if elem["lexicalized"] == "false" else True
        if elem.get("subcat"):
            elem["subcat"] = elem["subcat"].split()


def _validate_frames(elems: List[_Elem]) -> None:
    for elem in elems:
        assert "subcategorizationFrame" in elem
        if elem.get("senses"):
            elem["senses"] = elem["senses"].split()


def _validate_synsets(elems: List[_Elem], extension: bool) -> None:
    for elem in elems:
        assert "id" in elem
        if not extension:
            assert not elem.get("external")
        if not elem.get("external"):
            assert "ili" in elem
            elem.setdefault("meta")
        for defn in elem.get("definitions", []):
            defn.setdefault("text", "")
            defn.setdefault("meta")
        for rel in elem.get("relations", []):
            assert "target" in rel
            assert "relType" in rel
            rel.setdefault("meta")
        for ex in elem.get("examples", []):
            ex.setdefault("text", "")
            ex.setdefault("meta")
        if elem.get("lexicalized"):
            elem["lexicalized"] = False if elem["lexicalized"] == "false" else True
        if elem.get("members"):
            elem["members"] = elem["members"].split()


def _validate_metadata(elem: _Elem) -> None:
    if elem.get("confidenceScore"):
        elem["confidenceScore"] = float(elem["confidenceScore"])


# Serialization ########################################################


def dump(resource: LexicalResource, destination: AnyPath) -> None:
    """Write wordnets in the WN-LMF format.

    Args:
        lexicons: a list of :class:`Lexicon` objects
    """
    version = resource["lmf_version"]
    if version not in SUPPORTED_VERSIONS:
        raise LMFError(f"invalid version: {version}")
    destination = Path(destination).expanduser()
    doctype = _DOCTYPE.format(schema=_SCHEMAS[version])
    dc_uri = _DC_URIS[version]
    _version = version_info(version)
    with destination.open("wt", encoding="utf-8") as out:
        print(_XMLDECL.decode("utf-8"), file=out)
        print(doctype, file=out)
        print(f'<LexicalResource xmlns:dc="{dc_uri}">', file=out)
        for lexicon in resource["lexicons"]:
            _dump_lexicon(lexicon, out, _version)
        print("</LexicalResource>", file=out)


def _dump_lexicon(
    lexicon: Union[Lexicon, LexiconExtension], out: TextIO, version: VersionInfo
) -> None:
    lexicontype = "LexiconExtension" if lexicon.get("extends") else "Lexicon"
    attrib = _build_lexicon_attrib(lexicon, version)
    attrdelim = "\n" + (" " * len(f"  <{lexicontype} "))
    attrs = attrdelim.join(
        f"{attr}={quoteattr(str(val))}" for attr, val in attrib.items()
    )
    print(f"  <{lexicontype} {attrs}>", file=out)

    if version >= (1, 1):
        if lexicontype == "LexiconExtension":
            assert lexicon.get("extends")
            lexicon = cast(LexiconExtension, lexicon)
            _dump_dependency(lexicon["extends"], "Extends", out)
        for req in lexicon.get("requires", []):
            _dump_dependency(req, "Requires", out)

    for entry in lexicon.get("entries", []):
        _dump_lexical_entry(entry, out, version)

    for synset in lexicon.get("synsets", []):
        _dump_synset(synset, out, version)

    if version >= (1, 1):
        for sb in lexicon.get("frames", []):
            _dump_syntactic_behaviour(sb, out, version)

    print(f"  </{lexicontype}>", file=out)


def _build_lexicon_attrib(
    lexicon: Union[Lexicon, LexiconExtension], version: VersionInfo
) -> Dict[str, str]:
    attrib = {
        "id": lexicon["id"],
        "label": lexicon["label"],
        "language": lexicon["language"],
        "email": lexicon["email"],
        "license": lexicon["license"],
        "version": lexicon["version"],
    }
    if lexicon.get("url"):
        attrib["url"] = lexicon["url"]
    if lexicon.get("citation"):
        attrib["citation"] = lexicon["citation"]
    if version >= (1, 1) and lexicon.get("logo"):
        attrib["logo"] = lexicon["logo"]
    attrib.update(_meta_dict(lexicon.get("meta")))
    return attrib


def _dump_dependency(dep: Dependency, deptype: str, out: TextIO) -> None:
    attrib = {"id": dep["id"], "version": dep["version"]}
    if dep.get("url"):
        attrib["url"] = dep["url"]
    elem = ET.Element(deptype, attrib=attrib)
    print(_tostring(elem, 2), file=out)


def _dump_lexical_entry(
    entry: Union[LexicalEntry, ExternalLexicalEntry],
    out: TextIO,
    version: VersionInfo,
) -> None:
    frames = []
    attrib = {"id": entry["id"]}
    if entry.get("external", False):
        elem = ET.Element("ExternalLexicalEntry", attrib=attrib)
        if entry.get("lemma"):
            assert entry["lemma"].get("external", False)
            elem.append(_build_lemma(entry["lemma"], version))
    else:
        entry = cast(LexicalEntry, entry)
        attrib.update(_meta_dict(entry.get("meta")))
        elem = ET.Element("LexicalEntry", attrib=attrib)
        elem.append(_build_lemma(entry["lemma"], version))
        if version < (1, 1):
            frames = [
                _build_syntactic_behaviour(sb, version)
                for sb in entry.get("frames", [])
            ]
    elem.extend([_build_form(form, version) for form in entry.get("forms", [])])
    elem.extend([_build_sense(sense, version) for sense in entry.get("senses", [])])
    elem.extend(frames)
    print(_tostring(elem, 2), file=out)


def _build_lemma(
    lemma: Union[Lemma, ExternalLemma], version: VersionInfo
) -> ET.Element:
    if lemma.get("external", False):
        elem = ET.Element("ExternalLemma")
    else:
        lemma = cast(Lemma, lemma)
        attrib = {"writtenForm": lemma["writtenForm"]}
        if lemma.get("script"):
            attrib["script"] = lemma["script"]
        attrib["partOfSpeech"] = lemma["partOfSpeech"]
        elem = ET.Element("Lemma", attrib=attrib)
    if version >= (1, 1):
        for pron in lemma.get("pronunciations", []):
            elem.append(_build_pronunciation(pron))
    for tag in lemma.get("tags", []):
        elem.append(_build_tag(tag))
    return elem


def _build_form(form: Union[Form, ExternalForm], version: VersionInfo) -> ET.Element:
    attrib = {}
    if version >= (1, 1) and form["id"]:
        attrib["id"] = form["id"]
    if form.get("external", False):
        elem = ET.Element("ExternalForm", attrib=attrib)
    else:
        form = cast(Form, form)
        attrib["writtenForm"] = form["writtenForm"]
        if form.get("script"):
            attrib["script"] = form["script"]
        elem = ET.Element("Form", attrib=attrib)
    if version >= (1, 1):
        for pron in form.get("pronunciations", []):
            elem.append(_build_pronunciation(pron))
    for tag in form.get("tags", []):
        elem.append(_build_tag(tag))
    return elem


def _build_pronunciation(pron: Pronunciation) -> ET.Element:
    attrib = {}
    if pron.get("variety"):
        attrib["variety"] = pron["variety"]
    if pron.get("notation"):
        attrib["notation"] = pron["notation"]
    if not pron.get("phonemic", True):
        attrib["phonemic"] = "false"
    if pron.get("audio"):
        attrib["audio"] = pron["audio"]
    elem = ET.Element("Pronunciation", attrib=attrib)
    elem.text = pron["text"]
    return elem


def _build_tag(tag: Tag) -> ET.Element:
    elem = ET.Element("Tag", category=tag["category"])
    elem.text = tag["text"]
    return elem


def _build_sense(
    sense: Union[Sense, ExternalSense],
    version: VersionInfo,
) -> ET.Element:
    attrib = {"id": sense["id"]}
    if sense.get("external"):
        elem = ET.Element("ExternalSense", attrib=attrib)
    else:
        sense = cast(Sense, sense)
        attrib["synset"] = sense["synset"]
        attrib.update(_meta_dict(sense.get("meta")))
        if not sense.get("lexicalized", True):
            attrib["lexicalized"] = "false"
        if sense.get("adjposition"):
            attrib["adjposition"] = sense["adjposition"]
        if version >= (1, 1) and sense.get("subcat"):
            attrib["subcat"] = " ".join(sense["subcat"])
        elem = ET.Element("Sense", attrib=attrib)
    elem.extend(
        [_build_relation(rel, "SenseRelation") for rel in sense.get("relations", [])]
    )
    elem.extend([_build_example(ex) for ex in sense.get("examples", [])])
    elem.extend([_build_count(cnt) for cnt in sense.get("counts", [])])
    return elem


def _build_example(example: Example) -> ET.Element:
    elem = ET.Element("Example")
    elem.text = example["text"]
    if example.get("language"):
        elem.set("language", example["language"])
    return elem


def _build_count(count: Count) -> ET.Element:
    elem = ET.Element("Count", attrib=_meta_dict(count.get("meta")))
    elem.text = str(count["value"])
    return elem


def _dump_synset(
    synset: Union[Synset, ExternalSynset], out: TextIO, version: VersionInfo
) -> None:
    attrib: Dict[str, str] = {"id": synset["id"]}
    if synset.get("external", False):
        elem = ET.Element("ExternalSynset", attrib=attrib)
        elem.extend([_build_definition(defn) for defn in synset.get("definitions", [])])
    else:
        synset = cast(Synset, synset)
        attrib["ili"] = synset["ili"]
        if synset.get("partOfSpeech"):
            attrib["partOfSpeech"] = synset["partOfSpeech"]
        if not synset.get("lexicalized", True):
            attrib["lexicalized"] = "false"
        if version >= (1, 1):
            if synset.get("members"):
                attrib["members"] = " ".join(synset["members"])
            if synset.get("lexfile"):
                attrib["lexfile"] = synset["lexfile"]
        attrib.update(_meta_dict(synset.get("meta")))
        elem = ET.Element("Synset", attrib=attrib)
        elem.extend([_build_definition(defn) for defn in synset.get("definitions", [])])
        if synset.get("ili_definition"):
            elem.append(_build_ili_definition(synset["ili_definition"]))
    elem.extend(
        [_build_relation(rel, "SynsetRelation") for rel in synset.get("relations", [])]
    )
    elem.extend([_build_example(ex) for ex in synset.get("examples", [])])
    print(_tostring(elem, 2), file=out)


def _build_definition(definition: Definition) -> ET.Element:
    attrib = {}
    if definition.get("language"):
        attrib["language"] = definition["language"]
    if definition.get("sourceSense"):
        attrib["sourceSense"] = definition["sourceSense"]
    attrib.update(_meta_dict(definition.get("meta")))
    elem = ET.Element("Definition", attrib=attrib)
    elem.text = definition["text"]
    return elem


def _build_ili_definition(ili_definition: ILIDefinition) -> ET.Element:
    elem = ET.Element("ILIDefinition", attrib=_meta_dict(ili_definition.get("meta")))
    elem.text = ili_definition["text"]
    return elem


def _build_relation(relation: Relation, elemtype: str) -> ET.Element:
    attrib = {"target": relation["target"], "relType": relation["relType"]}
    attrib.update(_meta_dict(relation.get("meta")))
    return ET.Element(elemtype, attrib=attrib)


def _dump_syntactic_behaviour(
    syntactic_behaviour: SyntacticBehaviour, out: TextIO, version: VersionInfo
) -> None:
    elem = _build_syntactic_behaviour(syntactic_behaviour, version)
    print("    " + _tostring(elem, 2), file=out)


def _build_syntactic_behaviour(
    syntactic_behaviour: SyntacticBehaviour, version: VersionInfo
) -> ET.Element:
    attrib = {"subcategorizationFrame": syntactic_behaviour["subcategorizationFrame"]}
    if version >= (1, 1) and syntactic_behaviour.get("id"):
        attrib["id"] = syntactic_behaviour["id"]
    elif version < (1, 1) and syntactic_behaviour.get("senses"):
        attrib["senses"] = " ".join(syntactic_behaviour["senses"])
    return ET.Element("SyntacticBehaviour", attrib=attrib)


def _tostring(elem: ET.Element, level: int, short_empty_elements: bool = True) -> str:
    _indent(elem, level)
    return ("  " * level) + ET.tostring(
        elem, encoding="unicode", short_empty_elements=short_empty_elements
    )


def _indent(elem: ET.Element, level: int) -> None:
    self_indent = "\n" + "  " * level
    child_indent = self_indent + "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = child_indent
        for child in elem[:-1]:
            _indent(child, level + 1)
            child.tail = child_indent
        _indent(elem[-1], level + 1)
        elem[-1].tail = self_indent


def _meta_dict(meta: Optional[Metadata]) -> Dict[str, str]:
    if meta is not None:
        # Literal keys are required for typing purposes, so first
        # construct the dict and then remove those that weren't specified.
        d = {
            "dc:contributor": meta.get("contributor", ""),
            "dc:coverage": meta.get("coverage", ""),
            "dc:creator": meta.get("creator", ""),
            "dc:date": meta.get("date", ""),
            "dc:description": meta.get("description", ""),
            "dc:format": meta.get("format", ""),
            "dc:identifier": meta.get("identifier", ""),
            "dc:publisher": meta.get("publisher", ""),
            "dc:relation": meta.get("relation", ""),
            "dc:rights": meta.get("rights", ""),
            "dc:source": meta.get("source", ""),
            "dc:subject": meta.get("subject", ""),
            "dc:title": meta.get("title", ""),
            "dc:type": meta.get("type", ""),
            "status": meta.get("status", ""),
            "note": meta.get("note", ""),
        }
        d = {key: val for key, val in d.items() if val}
        # this one requires a conversion, so do it separately
        if "confidenceScore" in meta:
            d["confidenceScore"] = str(meta["confidenceScore"])
    else:
        d = {}
    return d



import sys
import argparse
from pathlib import Path
import json
import logging

import wn
from wn.project import iterpackages
from wn import lmf
from wn.validate import validate


def _download(args):
    if args.index:
        wn.config.load_index(args.index)
    for target in args.target:
        wn.download(target, add=args.add)


def _lexicons(args):
    for lex in wn.lexicons(lang=args.lang, lexicon=args.lexicon):
        print("\t".join((lex.id, lex.version, f"[{lex.language}]", lex.label)))


def _projects(args):
    for info in wn.projects():
        key = "i"
        key += "c" if info["cache"] else "-"
        # key += 'a' if False else '-'  # TODO: check if project is added to db
        print(
            "\t".join(
                (
                    key,
                    info["id"],
                    info["version"],
                    f"[{info['language'] or '---'}]",
                    info["label"] or "---",
                )
            )
        )


def _validate(args):
    all_valid = True
    selectseq = [check.strip() for check in args.select.split(",")]
    for package in iterpackages(args.FILE):
        resource = lmf.load(package.resource_file())
        for lexicon in resource["lexicons"]:
            spec = f'{lexicon["id"]}:{lexicon["version"]}'
            print(f"{spec:<20}", end="")
            report = validate(lexicon, select=selectseq)
            if not any(check.get("items", []) for check in report.values()):
                print("passed")
            else:
                print("failed")
                all_valid = False
                # clean up report
                for code in list(report):
                    if not report[code].get("items"):
                        del report[code]
                if args.output_file:
                    with open(args.output_file, "w") as outfile:
                        json.dump(report, outfile, indent=2)
                else:
                    for code, check in report.items():
                        if not check["items"]:
                            continue
                        print(f'  {check["message"]}')
                        for id, context in check["items"].items():
                            print(f"    {id}: {context}" if context else f"    {id}")

    sys.exit(0 if all_valid else 1)


def _path_type(arg):
    return Path(arg)


def _file_path_type(arg):
    path = Path(arg)
    if not path.is_file():
        raise argparse.ArgumentTypeError(f"cannot file file: {arg}")
    return path


parser = argparse.ArgumentParser(
    prog="python3 -m wn",
    description="Manage Wn's wordnet data from the command line.",
)
parser.add_argument("-V", "--version", action="version", version=f"Wn {wn.__version__}")
parser.add_argument(
    "-v",
    "--verbose",
    action="count",
    dest="verbosity",
    default=0,
    help="increase verbosity (can repeat: -vv, -vvv)",
)
parser.add_argument(
    "-d",
    "--dir",
    type=_path_type,
    help="data directory for Wn's database and cache",
)
parser.set_defaults(func=lambda _: parser.print_help())
sub_parsers = parser.add_subparsers(title="subcommands")


parser_download = sub_parsers.add_parser(
    "download",
    description="Download wordnets and add them to Wn's database.",
    help="download wordnets",
)
parser_download.add_argument("target", nargs="+", help="project specifiers or URLs")
parser_download.add_argument(
    "--index", type=_file_path_type, help="project index to use for downloading"
)
parser_download.add_argument(
    "--no-add",
    action="store_false",
    dest="add",
    help="download and cache without adding to the database",
)
parser_download.set_defaults(func=_download)


parser_lexicons = sub_parsers.add_parser(
    "lexicons",
    description="Display a list of installed lexicons.",
    help="list installed lexicons",
)
parser_lexicons.add_argument("-l", "--lang", help="BCP 47 language code")
parser_lexicons.add_argument("--lexicon", help="lexicon specifiers")
parser_lexicons.set_defaults(func=_lexicons)


parser_projects = sub_parsers.add_parser(
    "projects",
    description=(
        "Display a list of known projects. The first column shows the "
        "status for a project (i=indexed, c=cached)."
    ),
    help="list known projects",
)
parser_projects.set_defaults(func=_projects)


parser_validate = sub_parsers.add_parser(
    "validate",
    description=("Validate a WN-LMF lexicon"),
    help="validate a lexicon",
)
parser_validate.add_argument(
    "FILE", type=_file_path_type, help="WN-LMF (XML) lexicon file to validate"
)
parser_validate.add_argument(
    "--select",
    metavar="CHECKS",
    default="E,W",
    help="comma-separated list of checks to run (default: E,W)",
)
parser_validate.add_argument(
    "--output-file", metavar="FILE", help="write report to a JSON file"
)
parser_validate.set_defaults(func=_validate)


args = parser.parse_args()

logging.basicConfig(level=logging.ERROR - (min(args.verbosity, 3) * 10))

if args.dir:
    wn.config.data_directory = args.dir

args.func(args)



"""
Storage back-end interface.
"""

from typing import Dict
from pathlib import Path
import json
import sqlite3
import logging

import wn
from wn._types import AnyPath
from wn._util import resources, short_hash


logger = logging.getLogger("wn")


# Module Constants

DEBUG = False
NON_ROWID = 0  # imaginary rowid of non-existent row

# This stores hashes of the schema to check for version differences.
# When the schema changes, the hash will change. If the new hash is
# not added here, the 'test_schema_compatibility' test will fail. It
# is the developer's responsibility to only add compatible schema
# hashes here. If the schema change is not backwards-compatible, then
# clear all old hashes and only put the latest hash here. A hash can
# be generated like this:
#
# >>> import sqlite3
# >>> import wn
# >>> conn = sqlite3.connect(wn.config.database_path)
# >>> wn._db.schema_hash(conn)
#
COMPATIBLE_SCHEMA_HASHES = {
    "4c8ad03af5422d6979039ee2b80838d07c12d2c8",
    "2ec5828f109abdcd61954bb967ec64324a8632f1",
}


# Optional metadata is stored as a JSON string


def _adapt_dict(d: dict) -> bytes:
    return json.dumps(d).encode("utf-8")


def _convert_dict(s: bytes) -> dict:
    return json.loads(s)


def _convert_boolean(s: bytes) -> bool:
    return bool(int(s))


sqlite3.register_adapter(dict, _adapt_dict)
sqlite3.register_converter("meta", _convert_dict)
sqlite3.register_converter("boolean", _convert_boolean)


# The pool is a cache of open connections. Unless the database path is
# changed, there should only be zero or one.
pool: Dict[AnyPath, sqlite3.Connection] = {}


# The connect() function should be used for all connections


def connect() -> sqlite3.Connection:
    dbpath = wn.config.database_path
    if dbpath not in pool:
        initialized = dbpath.is_file()
        conn = sqlite3.connect(
            str(dbpath),
            detect_types=sqlite3.PARSE_DECLTYPES,
            check_same_thread=not wn.config.allow_multithreading,
        )
        # foreign key support needs to be enabled for each connection
        conn.execute("PRAGMA foreign_keys = ON")
        if DEBUG:
            conn.set_trace_callback(print)
        if not initialized:
            logger.info("initializing database: %s", dbpath)
            _init_db(conn)
        _check_schema_compatibility(conn, dbpath)

        pool[dbpath] = conn
    return pool[dbpath]


def _init_db(conn: sqlite3.Connection) -> None:
    schema = resources.read_text("wn", "schema.sql")
    conn.executescript(schema)
    with conn:
        conn.executemany(
            "INSERT INTO ili_statuses VALUES (null,?)",
            [("presupposed",), ("proposed",)],
        )


def _check_schema_compatibility(conn: sqlite3.Connection, dbpath: Path) -> None:
    hash = schema_hash(conn)

    # if the hash is known, then we're all good here
    if hash in COMPATIBLE_SCHEMA_HASHES:
        return

    logger.debug("current schema hash:\n  %s", hash)
    logger.debug(
        "compatible schema hashes:\n  %s", "\n  ".join(COMPATIBLE_SCHEMA_HASHES)
    )
    # otherwise, try to raise a helpful error message
    msg = (
        "Wn's schema has changed and is no longer compatible with the "
        f"database. Please move or delete {dbpath} and rebuild it."
    )
    try:
        specs = conn.execute("SELECT id, version FROM lexicons").fetchall()
    except sqlite3.OperationalError as exc:
        raise wn.DatabaseError(msg) from exc
    else:
        if specs:
            installed = "\n  ".join(f"{id}:{ver}" for id, ver in specs)
            msg += f" Lexicons currently installed:\n  {installed}"
        else:
            msg += " No lexicons are currently installed."
        raise wn.DatabaseError(msg)


def schema_hash(conn: sqlite3.Connection) -> str:
    query = "SELECT sql FROM sqlite_master WHERE NOT sql ISNULL"
    schema = "\n\n".join(row[0] for row in conn.execute(query))
    return short_hash(schema)



"""
Database retrieval queries.
"""
import asyncio
from typing import Optional, List, Tuple, Collection, Iterator, Sequence
import itertools
import sqlite3

import wn
from wn import DatabaseError
from wn._types import Metadata
from wn._db import connect, NON_ROWID, schema_hash

# Local Types

_Pronunciation = Tuple[
    str,  # value
    str,  # variety
    str,  # notation
    bool,  # phonemic
    str,  # audio
]
_Tag = Tuple[str, str]  # tag, category
_Form = Tuple[str, Optional[str], Optional[str], int]  # form  # id  # script  # rowid
_Word = Tuple[
    str,  # id
    str,  # pos
    List[_Form],  # forms
    int,  # lexid
    int,  # rowid
]
_Synset = Tuple[
    str,  # id
    str,  # pos
    str,  # ili
    int,  # lexid
    int,  # rowid
]
_Synset_Relation = Tuple[str, int, str, str, str, int, int]  # relname, relid, *_Synset
_Sense = Tuple[
    str,  # id
    str,  # entry_id
    str,  # synset_id
    int,  # lexid
    int,  # rowid
]
_Sense_Relation = Tuple[str, int, str, str, str, int, int]  # relname, relid,  *_Sense
_Count = Tuple[int, int]  # count, count_id
_SyntacticBehaviour = Tuple[str, str, List[str]]  # id  # frame  # sense ids
_ILI = Tuple[
    Optional[str],  # id
    str,  # status
    Optional[str],  # definition
    int,  # rowid
]
_Lexicon = Tuple[
    int,  # rowid
    str,  # id
    str,  # label
    str,  # language
    str,  # email
    str,  # license
    str,  # version
    str,  # url
    str,  # citation
    str,  # logo
    Metadata,  # metadata
]

refreshing = False
created = False


async def refresh_keyword_matching_table(keywords: List[str]) -> None:
    if not keywords:
        return
    global created
    global refreshing
    refreshing = True
    if not wn.config.allow_multithreading:
        print("Hi")
        raise DatabaseError("Can not refresh table since multithreading is disabled")
    else:
        conn = connect()
        cur = conn.cursor()
        q1 = """
        DROP TABLE IF EXISTS keyword_matches
        """
        q2 = f"""
        CREATE TABLE keyword_matches AS
WITH sy AS (SELECT DISTINCT s3.synset_rowid
            FROM forms f
                     JOIN senses s ON f.entry_rowid = s.entry_rowid
                     JOIN senses s1 ON s.synset_rowid = s1.synset_rowid
                     JOIN forms f2 ON s1.entry_rowid = f2.entry_rowid
                     JOIN senses s3 ON f2.entry_rowid = s3.entry_rowid),
     graph AS (SELECT source_rowid AS sid1, target_rowid AS sid2, source_rowid AS start
               FROM synsets
                        JOIN synset_relations sr ON synsets.rowid = sr.source_rowid
                        JOIN relation_types rt ON sr.type_rowid = rt.rowid
               WHERE type = 'hypernym'
                 AND source_rowid IN sy),
     relations AS (SELECT sid1, sid2, start
                   FROM graph
                   UNION ALL
                   SELECT sr.source_rowid AS sid1, sr.target_rowid AS sid2, start
                   FROM relations rvr
                            JOIN synset_relations sr ON rvr.sid2 = sr.source_rowid
                            JOIN relation_types r ON sr.type_rowid = r.rowid
                   WHERE type = 'hypernym'),
     fr AS (SELECT f2.form, relations.start
            FROM relations
                     JOIN senses ON relations.sid2 = senses.synset_rowid
                     JOIN forms f2 ON senses.entry_rowid = f2.entry_rowid)
SELECT DISTINCT f3.form AS word, fr.form AS keyword
FROM fr
         JOIN senses ON senses.synset_rowid = fr.start
         JOIN forms f3 ON senses.entry_rowid = f3.entry_rowid
WHERE fr.form IN ({_qs(keywords)});
        """
        q3 = """
CREATE INDEX k_index ON keyword_matches (word)
        """
        cur.execute(q1)
        cur.execute(q2, keywords)
        cur.execute(q3)
        conn.commit()
        print("ok")
        created = True
        refreshing = False


def match_for_keyword_in_hypernym_graph(
    word: str, keywords: List[str] = None, cur=None, recompute_with_new_keywords=False
) -> Iterator[Tuple[str]]:
    global refreshing
    global created
    if not keywords:
        keywords = wn.config.match_on_keywords
    if not keywords:
        return
    if cur is None:
        cur = connect().cursor()
    if (
        not created
        and (keywords == wn.config.match_on_keywords or recompute_with_new_keywords)
        and not refreshing
    ):
        refreshing = True
        asyncio.get_event_loop().create_task(refresh_keyword_matching_table(keywords))
        print(schema_hash(connect()))
    if not refreshing and keywords == wn.config.match_on_keywords:
        query = """
        SELECT keyword FROM keyword_matches WHERE word = ?
        """
        for row in cur.execute(query, [word]):
            yield row
        return
    query = f"""
    WITH sy AS (SELECT DISTINCT s3.synset_rowid
            FROM forms f
                     JOIN senses s ON f.entry_rowid = s.entry_rowid
                     JOIN senses s1 ON s.synset_rowid = s1.synset_rowid
                     JOIN forms f2 ON s1.entry_rowid = f2.entry_rowid
                     JOIN senses s3 ON f2.entry_rowid = s3.entry_rowid
            WHERE f2.form = ?),
     graph AS (SELECT source_rowid AS sid1, target_rowid AS sid2
               FROM synsets
                        JOIN synset_relations sr ON synsets.rowid = sr.source_rowid
                        JOIN relation_types rt ON sr.type_rowid = rt.rowid
               WHERE type = 'hypernym'
                 AND source_rowid IN sy),
     relations AS (SELECT sid1, sid2
                   FROM graph
                   UNION ALL
                   SELECT sr.source_rowid AS sid1, sr.target_rowid AS sid2
                   FROM relations rvr
                            JOIN synset_relations sr ON rvr.sid2 = sr.source_rowid
                            JOIN relation_types r ON sr.type_rowid = r.rowid
                   WHERE type = 'hypernym'),
     fr AS (SELECT f2.form
            FROM relations
                     JOIN senses ON relations.sid2 = senses.synset_rowid
                     JOIN forms f2 ON senses.entry_rowid = f2.entry_rowid)
SELECT DISTINCT fr.form AS keyword
FROM fr
WHERE fr.form IN ({_qs(keywords)})
    """
    for row in cur.execute(query, ([word] + keywords)):
        yield row


def find_lexicons(
    lexicon: str,
    lang: str = None,
) -> Iterator[_Lexicon]:
    cur = connect().cursor()
    found = False
    for specifier in lexicon.split():
        limit = "-1" if "*" in lexicon else "1"
        if ":" not in specifier:
            specifier += ":*"
        query = f"""
            SELECT DISTINCT rowid, id, label, language, email, license,
                            version, url, citation, logo
              FROM lexicons
             WHERE id || ":" || version GLOB :specifier
               AND (:language ISNULL OR language = :language)
             LIMIT {limit}
        """
        params = {"specifier": specifier, "language": lang}
        for row in cur.execute(query, params):
            yield row
            found = True
    # only raise an error when the query specifies something
    if not found and (lexicon != "*" or lang is not None):
        raise wn.Error(f"no lexicon found with lang={lang!r} and lexicon={lexicon!r}")


def get_lexicon(rowid: int) -> _Lexicon:
    conn = connect()
    return _get_lexicon(conn, rowid)


def _get_lexicon(conn: sqlite3.Connection, rowid: int) -> _Lexicon:
    query = """
        SELECT DISTINCT rowid, id, label, language, email, license,
                        version, url, citation, logo
        FROM lexicons
        WHERE rowid = ?
    """
    row: Optional[_Lexicon] = conn.execute(query, (rowid,)).fetchone()
    if row is None:
        raise LookupError(rowid)  # should we have a WnLookupError?
    return row


def get_modified(rowid: int) -> bool:
    query = "SELECT modified FROM lexicons WHERE rowid = ?"
    return connect().execute(query, (rowid,)).fetchone()[0]


def get_lexicon_dependencies(rowid: int) -> List[Tuple[str, str, str, Optional[int]]]:
    query = """
        SELECT provider_id, provider_version, provider_url, provider_rowid
          FROM lexicon_dependencies
         WHERE dependent_rowid = ?
    """
    return connect().execute(query, (rowid,)).fetchall()


def get_lexicon_extension_bases(rowid: int, depth: int = -1) -> List[int]:
    query = """
          WITH RECURSIVE ext(x, d) AS
               (SELECT base_rowid, 1
                  FROM lexicon_extensions
                 WHERE extension_rowid = :rowid
                 UNION SELECT base_rowid, d+1
                         FROM lexicon_extensions
                         JOIN ext ON extension_rowid = x)
        SELECT x FROM ext
         WHERE :depth < 0 OR d <= :depth
         ORDER BY d
    """
    rows = connect().execute(query, {"rowid": rowid, "depth": depth})
    return [row[0] for row in rows]


def get_lexicon_extensions(rowid: int, depth: int = -1) -> List[int]:
    query = """
          WITH RECURSIVE ext(x, d) AS
               (SELECT extension_rowid, 1
                  FROM lexicon_extensions
                 WHERE base_rowid = :rowid
                 UNION SELECT extension_rowid, d+1
                         FROM lexicon_extensions
                         JOIN ext ON base_rowid = x)
        SELECT x FROM ext
         WHERE :depth < 0 OR d <= :depth
         ORDER BY d
    """
    rows = connect().execute(query, {"rowid": rowid, "depth": depth})
    return [row[0] for row in rows]


def find_ilis(
    id: str = None,
    status: str = None,
    lexicon_rowids: Sequence[int] = None,
) -> Iterator[_ILI]:
    if status != "proposed":
        yield from _find_existing_ilis(
            id=id, status=status, lexicon_rowids=lexicon_rowids
        )
    if not id and (not status or status == "proposed"):
        yield from find_proposed_ilis(lexicon_rowids=lexicon_rowids)


def _find_existing_ilis(
    id: str = None,
    status: str = None,
    lexicon_rowids: Sequence[int] = None,
) -> Iterator[_ILI]:
    query = """
        SELECT DISTINCT i.id, ist.status, i.definition, i.rowid
          FROM ilis AS i
          JOIN ili_statuses AS ist ON i.status_rowid = ist.rowid
    """
    conditions: List[str] = []
    params: List = []
    if id:
        conditions.append("i.id = ?")
        params.append(id)
    if status:
        conditions.append("ist.status = ?")
        params.append(status)
    if lexicon_rowids:
        conditions.append(
            f"""
            i.rowid IN
            (SELECT ss.ili_rowid
               FROM synsets AS ss
              WHERE ss.lexicon_rowid IN ({_qs(lexicon_rowids)}))
        """
        )
        params.extend(lexicon_rowids)
    if conditions:
        query += "\n WHERE " + "\n   AND ".join(conditions)
    yield from connect().execute(query, params)


def find_proposed_ilis(
    synset_rowid: int = None,
    lexicon_rowids: Sequence[int] = None,
) -> Iterator[_ILI]:
    query = """
        SELECT null, "proposed", definition, rowid
          FROM proposed_ilis
    """
    conditions = []
    params = []
    if synset_rowid is not None:
        conditions.append("synset_rowid = ?")
        params.append(synset_rowid)
    if lexicon_rowids:
        conditions.append(
            f"""
            synset_rowid IN
            (SELECT ss.rowid FROM synsets AS ss
              WHERE ss.lexicon_rowid IN ({_qs(lexicon_rowids)}))
        """
        )
        params.extend(lexicon_rowids)
    if conditions:
        query += "\n WHERE " + "\n   AND ".join(conditions)
    yield from connect().execute(query, params)


def find_entries(
    id: str = None,
    forms: Sequence[str] = None,
    pos: str = None,
    lexicon_rowids: Sequence[int] = None,
    normalized: bool = False,
    search_all_forms: bool = False,
) -> Iterator[_Word]:
    conn = connect()
    cte = ""
    params: List = []
    conditions = []
    if id:
        conditions.append("e.id = ?")
        params.append(id)
    if forms:
        cte = f"WITH wordforms(s) AS (VALUES {_vs(forms)})"
        or_norm = "OR normalized_form IN wordforms" if normalized else ""
        and_rank = "" if search_all_forms else "AND rank = 0"
        conditions.append(
            f"""
            e.rowid IN
               (SELECT entry_rowid
                  FROM forms
                 WHERE (form IN wordforms {or_norm}) {and_rank})
        """.strip()
        )
        params.extend(forms)
    if pos:
        conditions.append("e.pos = ?")
        params.append(pos)
    if lexicon_rowids:
        conditions.append(f"e.lexicon_rowid IN ({_qs(lexicon_rowids)})")
        params.extend(lexicon_rowids)

    condition = ""
    if conditions:
        condition = "WHERE " + "\n           AND ".join(conditions)

    query = f"""
          {cte}
        SELECT DISTINCT e.lexicon_rowid, e.rowid, e.id, e.pos,
                        f.form, f.id, f.script, f.rowid
          FROM entries AS e
          JOIN forms AS f ON f.entry_rowid = e.rowid
         {condition}
         ORDER BY e.rowid, e.id, f.rank
    """

    rows: Iterator[
        Tuple[int, int, str, str, str, Optional[str], Optional[str], int]
    ] = conn.execute(query, params)
    groupby = itertools.groupby
    for key, group in groupby(rows, lambda row: row[0:4]):
        lexid, rowid, id, pos = key
        wordforms = [(row[4], row[5], row[6], row[7]) for row in group]
        yield (id, pos, wordforms, lexid, rowid)


def find_senses(
    id: str = None,
    forms: Sequence[str] = None,
    pos: str = None,
    lexicon_rowids: Sequence[int] = None,
    normalized: bool = False,
    search_all_forms: bool = False,
) -> Iterator[_Sense]:
    conn = connect()
    cte = ""
    params: List = []
    conditions = []
    if id:
        conditions.append("s.id = ?")
        params.append(id)
    if forms:
        cte = f"WITH wordforms(s) AS (VALUES {_vs(forms)})"
        or_norm = "OR normalized_form IN wordforms" if normalized else ""
        and_rank = "" if search_all_forms else "AND rank = 0"
        conditions.append(
            f"""
            s.entry_rowid IN
               (SELECT entry_rowid
                  FROM forms
                 WHERE (form IN wordforms {or_norm}) {and_rank})
        """.strip()
        )
        params.extend(forms)
    if pos:
        conditions.append("e.pos = ?")
        params.append(pos)
    if lexicon_rowids:
        conditions.append(f"s.lexicon_rowid IN ({_qs(lexicon_rowids)})")
        params.extend(lexicon_rowids)

    condition = ""
    if conditions:
        condition = "WHERE " + "\n           AND ".join(conditions)

    query = f"""
          {cte}
        SELECT DISTINCT s.id, e.id, ss.id, s.lexicon_rowid, s.rowid
          FROM senses AS s
          JOIN entries AS e ON e.rowid = s.entry_rowid
          JOIN synsets AS ss ON ss.rowid = s.synset_rowid
         {condition}
    """

    rows: Iterator[_Sense] = conn.execute(query, params)
    yield from rows


def find_synsets(
    id: str = None,
    forms: Sequence[str] = None,
    pos: str = None,
    ili: str = None,
    lexicon_rowids: Sequence[int] = None,
    normalized: bool = False,
    search_all_forms: bool = False,
) -> Iterator[_Synset]:
    conn = connect()
    cte = ""
    join = ""
    conditions = []
    order = ""
    params: List = []
    if id:
        conditions.append("ss.id = ?")
        params.append(id)
    if forms:
        cte = f"WITH wordforms(s) AS (VALUES {_vs(forms)})"
        or_norm = "OR normalized_form IN wordforms" if normalized else ""
        and_rank = "" if search_all_forms else "AND rank = 0"
        join = f"""\
          JOIN (SELECT _s.entry_rowid, _s.synset_rowid, _s.entry_rank
                  FROM forms AS f
                  JOIN senses AS _s ON _s.entry_rowid = f.entry_rowid
                 WHERE (f.form IN wordforms {or_norm}) {and_rank}) AS s
            ON s.synset_rowid = ss.rowid
        """.strip()
        params.extend(forms)
        order = "ORDER BY s.entry_rowid, s.entry_rank"
    if pos:
        conditions.append("ss.pos = ?")
        params.append(pos)
    if ili:
        conditions.append(
            "ss.ili_rowid IN (SELECT ilis.rowid FROM ilis WHERE ilis.id = ?)"
        )
        params.append(ili)
    if lexicon_rowids:
        conditions.append(f"ss.lexicon_rowid IN ({_qs(lexicon_rowids)})")
        params.extend(lexicon_rowids)

    condition = ""
    if conditions:
        condition = "WHERE " + "\n           AND ".join(conditions)

    query = f"""
          {cte}
        SELECT DISTINCT ss.id, ss.pos,
                        (SELECT ilis.id FROM ilis WHERE ilis.rowid=ss.ili_rowid),
                        ss.lexicon_rowid, ss.rowid
          FROM synsets AS ss
          {join}
         {condition}
         {order}
    """

    rows: Iterator[_Synset] = conn.execute(query, params)
    yield from rows


def get_synsets_for_ilis(
    ilis: Collection[str],
    lexicon_rowids: Sequence[int],
) -> Iterator[_Synset]:
    conn = connect()
    query = f"""
        SELECT DISTINCT ss.id, ss.pos, ili.id, ss.lexicon_rowid, ss.rowid
          FROM synsets as ss
          JOIN ilis as ili ON ss.ili_rowid = ili.rowid
         WHERE ili.id IN ({_qs(ilis)})
           AND ss.lexicon_rowid IN ({_qs(lexicon_rowids)})
    """
    params = *ilis, *lexicon_rowids
    result_rows: Iterator[_Synset] = conn.execute(query, params)
    yield from result_rows


def get_synset_relations(
    source_rowids: Collection[int],
    relation_types: Collection[str],
    lexicon_rowids: Sequence[int],
) -> Iterator[_Synset_Relation]:
    conn = connect()
    params: List = []
    constraint = ""
    if relation_types and "*" not in relation_types:
        constraint = f"WHERE type IN ({_qs(relation_types)})"
        params.extend(relation_types)
    params.extend(source_rowids)
    params.extend(lexicon_rowids)
    query = f"""
          WITH rt(rowid, type) AS
               (SELECT rowid, type FROM relation_types {constraint})
        SELECT DISTINCT rel.type, rel.rowid, tgt.id, tgt.pos,
                        (SELECT ilis.id FROM ilis WHERE ilis.rowid = tgt.ili_rowid),
                        tgt.lexicon_rowid, tgt.rowid
          FROM (SELECT rt.type, target_rowid, srel.rowid
                  FROM synset_relations AS srel
                  JOIN rt ON srel.type_rowid = rt.rowid
                 WHERE source_rowid IN ({_qs(source_rowids)})
                   AND lexicon_rowid IN ({_qs(lexicon_rowids)})
               ) AS rel
          JOIN synsets AS tgt
            ON tgt.rowid = rel.target_rowid
    """
    result_rows: Iterator[_Synset_Relation] = conn.execute(query, params)
    yield from result_rows


def get_definitions(
    synset_rowid: int,
    lexicon_rowids: Sequence[int],
) -> List[Tuple[str, str, str, int]]:
    conn = connect()
    query = f"""
        SELECT d.definition,
               d.language,
               (SELECT s.id FROM senses AS s WHERE s.rowid=d.sense_rowid),
               d.rowid
          FROM definitions AS d
         WHERE d.synset_rowid = ?
           AND d.lexicon_rowid IN ({_qs(lexicon_rowids)})
    """
    return conn.execute(query, (synset_rowid, *lexicon_rowids)).fetchall()


_SANITIZED_EXAMPLE_PREFIXES = {
    "senses": "sense",
    "synsets": "synset",
}


def get_examples(
    rowid: int,
    table: str,
    lexicon_rowids: Sequence[int],
) -> List[Tuple[str, str, int]]:
    conn = connect()
    prefix = _SANITIZED_EXAMPLE_PREFIXES.get(table)
    if prefix is None:
        raise wn.Error(f"'{table}' does not have examples")
    query = f"""
        SELECT example, language, rowid
          FROM {prefix}_examples
         WHERE {prefix}_rowid = ?
           AND lexicon_rowid IN ({_qs(lexicon_rowids)})
    """
    return conn.execute(query, (rowid, *lexicon_rowids)).fetchall()


def find_syntactic_behaviours(
    id: str = None,
    lexicon_rowids: Sequence[int] = None,
) -> Iterator[_SyntacticBehaviour]:
    conn = connect()
    query = """
        SELECT sb.id, sb.frame, s.id
          FROM syntactic_behaviours AS sb
          JOIN syntactic_behaviour_senses AS sbs
            ON sbs.syntactic_behaviour_rowid = sb.rowid
          JOIN senses AS s
            ON s.rowid = sbs.sense_rowid
    """
    conditions: List[str] = []
    params: List = []
    if id:
        conditions.append("sb.id = ?")
        params.append(id)
    if lexicon_rowids:
        conditions.append(f"sb.lexicon_rowid IN ({_qs(lexicon_rowids)})")
        params.extend(lexicon_rowids)
    if conditions:
        query += "\n WHERE " + "\n   AND ".join(conditions)
    rows: Iterator[Tuple[str, str, str]] = conn.execute(query, params)
    for key, group in itertools.groupby(rows, lambda row: row[0:2]):
        id, frame = key
        sense_ids = [row[2] for row in group]
        yield id, frame, sense_ids


def get_syntactic_behaviours(
    rowid: int,
    lexicon_rowids: Sequence[int],
) -> List[str]:
    conn = connect()
    query = f"""
        SELECT sb.frame
          FROM syntactic_behaviours AS sb
          JOIN syntactic_behaviour_senses AS sbs
            ON sbs.syntactic_behaviour_rowid = sb.rowid
         WHERE sbs.sense_rowid = ?
           AND sb.lexicon_rowid IN ({_qs(lexicon_rowids)})
    """
    return [row[0] for row in conn.execute(query, (rowid, *lexicon_rowids))]


def _get_senses(
    rowid: int,
    sourcetype: str,
    lexicon_rowids: Sequence[int],
) -> Iterator[_Sense]:
    conn = connect()
    query = f"""
        SELECT s.id, e.id, ss.id, s.lexicon_rowid, s.rowid
          FROM senses AS s
          JOIN entries AS e
            ON e.rowid = s.entry_rowid
          JOIN synsets AS ss
            ON ss.rowid = s.synset_rowid
         WHERE s.{sourcetype}_rowid = ?
           AND s.lexicon_rowid IN ({_qs(lexicon_rowids)})
         ORDER BY s.{sourcetype}_rank
    """
    return conn.execute(query, (rowid, *lexicon_rowids))


def get_entry_senses(rowid: int, lexicon_rowids: Sequence[int]) -> Iterator[_Sense]:
    yield from _get_senses(rowid, "entry", lexicon_rowids)


def get_synset_members(rowid: int, lexicon_rowids: Sequence[int]) -> Iterator[_Sense]:
    yield from _get_senses(rowid, "synset", lexicon_rowids)


def get_sense_relations(
    source_rowid: int,
    relation_types: Collection[str],
    lexicon_rowids: Sequence[int],
) -> Iterator[_Sense_Relation]:
    params: List = []
    constraint = ""
    if relation_types and "*" not in relation_types:
        constraint = f"WHERE type IN ({_qs(relation_types)})"
        params.extend(relation_types)
    params.append(source_rowid)
    params.extend(lexicon_rowids)
    query = f"""
          WITH rt(rowid, type) AS
               (SELECT rowid, type FROM relation_types {constraint})
        SELECT DISTINCT rel.type, rel.rowid,
                        s.id, e.id, ss.id,
                        s.lexicon_rowid, s.rowid
          FROM (SELECT rt.type, target_rowid, srel.rowid
                  FROM sense_relations AS srel
                  JOIN rt ON srel.type_rowid = rt.rowid
                 WHERE source_rowid = ?
                   AND lexicon_rowid IN ({_qs(lexicon_rowids)})
               ) AS rel
          JOIN senses AS s
            ON s.rowid = rel.target_rowid
          JOIN entries AS e
            ON e.rowid = s.entry_rowid
          JOIN synsets AS ss
            ON ss.rowid = s.synset_rowid
    """
    rows: Iterator[_Sense_Relation] = connect().execute(query, params)
    yield from rows


def get_sense_synset_relations(
    source_rowid: int,
    relation_types: Collection[str],
    lexicon_rowids: Sequence[int],
) -> Iterator[_Synset_Relation]:
    params: List = []
    constraint = ""
    if "*" not in relation_types:
        constraint = f"WHERE type IN ({_qs(relation_types)})"
        params.extend(relation_types)
    params.append(source_rowid)
    params.extend(lexicon_rowids)
    query = f"""
          WITH rt(rowid, type) AS
               (SELECT rowid, type FROM relation_types {constraint})
        SELECT DISTINCT rel.type, rel.rowid, ss.id, ss.pos,
                        (SELECT ilis.id FROM ilis WHERE ilis.rowid = ss.ili_rowid),
                        ss.lexicon_rowid, ss.rowid
          FROM (SELECT rt.type, target_rowid, srel.rowid
                  FROM sense_synset_relations AS srel
                  JOIN rt ON srel.type_rowid = rt.rowid
                 WHERE source_rowid = ?
                   AND lexicon_rowid IN ({_qs(lexicon_rowids)})
               ) AS rel
          JOIN synsets AS ss
            ON ss.rowid = rel.target_rowid
    """
    rows: Iterator[_Synset_Relation] = connect().execute(query, params)
    yield from rows


_SANITIZED_METADATA_TABLES = {
    "ilis": "ilis",
    "proposed_ilis": "proposed_ilis",
    "lexicons": "lexicons",
    "entries": "entries",
    "senses": "senses",
    "synsets": "synsets",
    "sense_relations": "sense_relations",
    "sense_synset_relations": "sense_synset_relations",
    "synset_relations": "synset_relations",
    "sense_examples": "sense_examples",
    "counts": "counts",
    "synset_examples": "synset_examples",
    "definitions": "definitions",
}


def get_metadata(rowid: int, table: str) -> Metadata:
    conn = connect()
    tablename = _SANITIZED_METADATA_TABLES.get(table)
    if tablename is None:
        raise wn.Error(f"'{table}' does not contain metadata")
    query = f"SELECT metadata FROM {tablename} WHERE rowid=?"
    return conn.execute(query, (rowid,)).fetchone()[0] or {}


_SANITIZED_LEXICALIZED_TABLES = {
    "senses": "senses",
    "synsets": "synsets",
}


def get_lexicalized(rowid: int, table: str) -> bool:
    conn = connect()
    tablename = _SANITIZED_LEXICALIZED_TABLES.get(table)
    if tablename is None:
        raise wn.Error(f"'{table}' does not mark lexicalization")
    if rowid == NON_ROWID:
        return False
    query = f"SELECT lexicalized FROM {tablename} WHERE rowid=?"
    return conn.execute(query, (rowid,)).fetchone()[0]


def get_adjposition(rowid: int) -> Optional[str]:
    conn = connect()
    query = "SELECT adjposition FROM adjpositions WHERE sense_rowid = ?"
    row = conn.execute(query, (rowid,)).fetchone()
    if row:
        return row[0]
    return None


def get_form_pronunciations(form_rowid: int) -> List[_Pronunciation]:
    # TODO: restrict by lexicon ids
    conn = connect()
    query = """
        SELECT value, variety, notation, phonemic, audio
          FROM pronunciations
         WHERE form_rowid = ?
    """
    rows: List[_Pronunciation] = conn.execute(query, (form_rowid,)).fetchall()
    return rows


def get_form_tags(form_rowid: int) -> List[_Tag]:
    # TODO: restrict by lexicon ids
    conn = connect()
    query = "SELECT tag, category FROM tags WHERE form_rowid = ?"
    rows: List[_Tag] = conn.execute(query, (form_rowid,)).fetchall()
    return rows


def get_sense_counts(sense_rowid: int, lexicon_rowids: Sequence[int]) -> List[_Count]:
    conn = connect()
    query = f"""
        SELECT count, rowid
          FROM counts
         WHERE sense_rowid = ?
           AND lexicon_rowid IN ({_qs(lexicon_rowids)})
    """
    rows: List[_Count] = conn.execute(query, (sense_rowid, *lexicon_rowids)).fetchall()
    return rows


def get_lexfile(synset_rowid: int) -> Optional[str]:
    conn = connect()
    query = """
        SELECT lf.name
          FROM lexfiles AS lf
          JOIN synsets AS ss ON ss.lexfile_rowid = lf.rowid
         WHERE ss.rowid = ?
    """
    row = conn.execute(query, (synset_rowid,)).fetchone()
    if row is not None and row[0] is not None:
        return row[0]
    return None


def _qs(xs: Collection) -> str:
    return ",".join("?" * len(xs))


def _vs(xs: Collection) -> str:
    return ",".join(["(?)"] * len(xs))


def _kws(xs: Collection) -> str:
    return ",".join(f":{x}" for x in xs)



"""
Wordnet Interface.
"""

__all__ = (
    "__version__",
    "Wordnet",
    "download",
    "add",
    "remove",
    "export",
    "projects",
    "lexicons",
    "Lexicon",
    "word",
    "words",
    "Word",
    "Form",
    "Pronunciation",
    "Tag",
    "sense",
    "senses",
    "Sense",
    "Count",
    "synset",
    "synsets",
    "Synset",
    "ili",
    "ilis",
    "ILI",
    "Error",
    "DatabaseError",
    "ConfigurationError",
    "ProjectError",
    "WnWarning",
)

from wn._exceptions import (
    Error,
    DatabaseError,
    ConfigurationError,
    ProjectError,
    WnWarning,
)
from wn._config import config  # noqa: F401
from wn._add import add, remove
from wn._export import export
from wn._download import download
from wn._core import (
    projects,
    lexicons,
    Lexicon,
    word,
    words,
    Word,
    Form,
    Pronunciation,
    Tag,
    sense,
    senses,
    Sense,
    Count,
    synset,
    synsets,
    Synset,
    ili,
    ilis,
    ILI,
    Wordnet,
)

__version__ = "0.9.3-5"



from typing import (
    Optional,
    Union,
    Callable,
    Mapping,
    Sequence,
    Tuple,
    Dict,
    Set,
    Any,
)
from pathlib import Path

# For functions taking a filesystem path as a str or a pathlib.Path
AnyPath = Union[str, Path]

# LMF versions for comparison
VersionInfo = Tuple[int, ...]

# Synset and Sense relations map a relation type to one or more ids
RelationMap = Mapping[str, Sequence[str]]

# User-facing metadata representation
Metadata = Dict[str, Any]

# A callable that returns a normalized word form for a given word form
NormalizeFunction = Callable[[str], str]

# Lemmatization returns a mapping of parts of speech (or None) to
# lists of wordforms that are potential lemmas for some query word
LemmatizeResult = Dict[Optional[str], Set[str]]

# A callable that returns a LemmatizationResult for a given word form
# and optional part of speech
LemmatizeFunction = Callable[[str, Optional[str]], LemmatizeResult]



from wn import util


def test_synset_id_formatter():
    f = util.synset_id_formatter
    assert f()(prefix="xyz", offset=123, pos="n") == "xyz-00000123-n"
    assert f(prefix="xyz")(offset=123, pos="n") == "xyz-00000123-n"
    assert f(prefix="xyz", pos="n")(offset=123) == "xyz-00000123-n"
    assert f("abc-{offset}-{pos}")(offset=1, pos="v") == "abc-1-v"



"""A simple English lemmatizer that finds and removes known suffixes.

"""

from typing import Optional, Dict, Set, List, Tuple
from enum import Flag, auto

import wn
from wn._types import LemmatizeResult
from wn.constants import NOUN, VERB, ADJ, ADJ_SAT, ADV, PARTS_OF_SPEECH

POSExceptionMap = Dict[str, Set[str]]
ExceptionMap = Dict[str, POSExceptionMap]


class _System(Flag):
    """Flags to track suffix rules in various implementations of Morphy."""

    PWN = auto()
    NLTK = auto()
    WN = auto()
    ALL = PWN | NLTK | WN


_PWN = _System.PWN
_NLTK = _System.NLTK
_WN = _System.WN
_ALL = _System.ALL


Rule = Tuple[str, str, _System]

DETACHMENT_RULES: Dict[str, List[Rule]] = {
    NOUN: [
        ("s", "", _ALL),
        ("ces", "x", _WN),
        ("ses", "s", _ALL),
        ("ves", "f", _NLTK | _WN),
        ("ives", "ife", _WN),
        ("xes", "x", _ALL),
        ("xes", "xis", _WN),
        ("zes", "z", _ALL),
        ("ches", "ch", _ALL),
        ("shes", "sh", _ALL),
        ("men", "man", _ALL),
        ("ies", "y", _ALL),
    ],
    VERB: [
        ("s", "", _ALL),
        ("ies", "y", _ALL),
        ("es", "e", _ALL),
        ("es", "", _ALL),
        ("ed", "e", _ALL),
        ("ed", "", _ALL),
        ("ing", "e", _ALL),
        ("ing", "", _ALL),
    ],
    ADJ: [
        ("er", "", _ALL),
        ("est", "", _ALL),
        ("er", "e", _ALL),
        ("est", "e", _ALL),
    ],
    ADV: [],
}
DETACHMENT_RULES[ADJ_SAT] = DETACHMENT_RULES[ADJ]


class Morphy:
    """The Morphy lemmatizer class.

    Objects of this class are callables that take a wordform and an
    optional part of speech and return a dictionary mapping parts of
    speech to lemmas. If objects of this class are not created with a
    :class:`wn.Wordnet` object, the returned lemmas may be invalid.

    Arguments:
        wordnet: optional :class:`wn.Wordnet` instance

    Example:

        >>> import wn
        >>> from wn.morphy import Morphy
        >>> ewn = wn.Wordnet('ewn:2020')
        >>> m = Morphy(ewn)
        >>> m('axes', pos='n')
        {'n': {'axe', 'ax', 'axis'}}
        >>> m('geese', pos='n')
        {'n': {'goose'}}
        >>> m('gooses')
        {'n': {'goose'}, 'v': {'goose'}}
        >>> m('goosing')
        {'v': {'goose'}}

    """

    def __init__(self, wordnet: Optional[wn.Wordnet] = None):
        self._rules = {
            pos: [rule for rule in rules if rule[2] & _System.WN]
            for pos, rules in DETACHMENT_RULES.items()
        }
        exceptions: ExceptionMap = {pos: {} for pos in PARTS_OF_SPEECH}
        all_lemmas: Dict[str, Set[str]] = {pos: set() for pos in PARTS_OF_SPEECH}
        if wordnet:
            for word in wordnet.words():
                pos = word.pos
                pos_exc = exceptions[pos]
                lemma, *others = word.forms()
                # store every lemma whether it has other forms or not
                all_lemmas[pos].add(lemma)
                # those with other forms map to the original lemmas
                for other in others:
                    if other in pos_exc:
                        pos_exc[other].add(lemma)
                    else:
                        pos_exc[other] = {lemma}
            self._initialized = True
        else:
            self._initialized = False
        self._exceptions = exceptions
        self._all_lemmas = all_lemmas

    def __call__(self, form: str, pos: Optional[str] = None) -> LemmatizeResult:
        result = {}
        if not self._initialized:
            result[pos] = {form}  # always include original when not initialized

        if pos is None:
            pos_list = list(DETACHMENT_RULES)
        elif pos in DETACHMENT_RULES:
            pos_list = [pos]
        else:
            pos_list = []  # not handled by morphy

        no_pos_forms = result.get(None, set())  # avoid unnecessary duplicates
        for _pos in pos_list:
            candidates = self._morphstr(form, _pos) - no_pos_forms
            if candidates:
                result.setdefault(_pos, set()).update(candidates)

        return result

    def _morphstr(self, form: str, pos: str) -> Set[str]:
        candidates: Set[str] = set()

        initialized = self._initialized
        if initialized:
            all_lemmas = self._all_lemmas[pos]
            if form in all_lemmas:
                candidates.add(form)
            candidates.update(self._exceptions[pos].get(form, set()))
        else:
            all_lemmas = set()

        for suffix, repl, _ in self._rules[pos]:
            # avoid applying rules that perform full suppletion
            if form.endswith(suffix) and len(suffix) < len(form):
                candidate = f"{form[:-len(suffix)]}{repl}"
                if not initialized or candidate in all_lemmas:
                    candidates.add(candidate)

        return candidates


morphy = Morphy()
